# pylint: skip-file
# type: ignore

import os
import pathlib as _pl
import shutil
import subprocess
import sys

from PyQt5.QtWidgets import *

import pytrnsys.utils.result as _res
import trnsysGUI.arguments as args
import trnsysGUI.buildDck as buildDck
import trnsysGUI.common.cancelled as _ccl
import trnsysGUI.diagram.Editor as _de
import trnsysGUI.images as _img
import trnsysGUI.project as _prj
import trnsysGUI.settings as _settings
import trnsysGUI.settingsDlg as _sdlg
import trnsysGUI.tracing as trc
from pytrnsys.utils import log
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.MassFlowVisualizer import MassFlowVisualizer
from trnsysGUI.ProcessMain import ProcessMain
from trnsysGUI.RunMain import RunMain
from trnsysGUI.configFile import configFile
from trnsysGUI.errors import showErrorMessageBox
from trnsysGUI.storageTank.widget import StorageTank

__version__ = "1.0.0"
__author__ = "Stefano Marti"
__email__ = "stefano.marti@spf.ch"
__status__ = "Prototype"


class _MainWindow(QMainWindow):
    def __init__(self, logger, project: _prj.Project, parent=None):
        super().__init__(parent)

        self.jsonPath = None
        self.logger = logger

        self.centralWidget = self._createDiagramEditor(project)
        self.setCentralWidget(self.centralWidget)

        self.labelVisState = False
        self.massFlowEnabled = False
        self.calledByVisualizeMf = False
        self.currentFile = "Untitled"

        # Toolbar actions
        saveDiaAction = QAction(_img.INBOX_PNG.icon(), "Save", self)
        saveDiaAction.triggered.connect(self.saveDia)

        loadDiaAction = QAction(_img.OUTBOX_PNG.icon(), "Open", self)
        loadDiaAction.triggered.connect(self.loadDia)

        updateConfigAction = QAction(
            _img.UPDATE_CONFIG_PNG.icon(), "Update run.config", self
        )
        updateConfigAction.triggered.connect(self.updateRun)

        runSimulationAction = QAction(
            _img.RUN_SIMULATION_PNG.icon(), "Run simulation", self
        )
        runSimulationAction.triggered.connect(self.runSimulation)

        processSimulationAction = QAction(
            _img.PROCESS_SIMULATION_PNG.icon(), "Process data", self
        )
        processSimulationAction.triggered.connect(self.processSimulation)

        deleteDiaAction = QAction(_img.TRASH_PNG.icon(), "Delete diagram", self)
        deleteDiaAction.triggered.connect(self.deleteDia)

        zoomInAction = QAction(_img.ZOOM_IN_PNG.icon(), "Zoom in", self)
        zoomInAction.triggered.connect(self.setZoomIn)

        zoomOutAction = QAction(_img.ZOOM_OUT_PNG.icon(), "Zoom out", self)
        zoomOutAction.triggered.connect(self.setZoomOut)

        toggleConnLabels = QAction(_img.LABEL_TOGGLE_PNG.icon(), "Toggle labels", self)
        toggleConnLabels.triggered.connect(self.toggleConnLabels)

        exportHydraulicsAction = QAction(
            _img.EXPORT_HYDRAULICS_PNG.icon(), "Export hydraulic.ddck", self
        )
        exportHydraulicsAction.triggered.connect(self.exportHydraulicsDdck)

        exportHydCtrlAction = QAction(
            _img.EXPORT_HYDRAULIC_CONTROL_PNG.icon(),
            "Export hydraulic_control.ddck",
            self,
        )
        exportHydCtrlAction.triggered.connect(self.exportHydraulicControl)

        exportDckAction = QAction(_img.EXPORT_DCK_PNG.icon(), "Export dck", self)
        exportDckAction.triggered.connect(self.exportDck)

        toggleSnapAction = QAction("Toggle snap grid", self)
        toggleSnapAction.triggered.connect(self.toggleSnap)
        toggleSnapAction.setShortcut("a")

        toggleAlignModeAction = QAction("Toggle align mode", self)
        toggleAlignModeAction.triggered.connect(self.toggleAlignMode)
        toggleAlignModeAction.setShortcut("q")

        runMassflowSolverAction = QAction(
            _img.RUN_MFS_PNG.icon(), "Run the massflow solver", self
        )
        runMassflowSolverAction.triggered.connect(self.runAndVisMf)

        openVisualizerAction = QAction(
            _img.VIS_MFS_PNG.icon(), "Start visualization of mass flows", self
        )
        openVisualizerAction.triggered.connect(self.visualizeMf)

        # Tool bar
        tb = self.addToolBar("Main Toolbar...")
        tb.setObjectName("Toolbar")
        tb.addAction(saveDiaAction)
        tb.addAction(loadDiaAction)
        tb.addAction(zoomInAction)
        tb.addAction(zoomOutAction)
        tb.addAction(toggleConnLabels)
        tb.addAction(runMassflowSolverAction)
        tb.addAction(openVisualizerAction)
        tb.addAction(exportHydraulicsAction)
        tb.addAction(exportHydCtrlAction)
        tb.addAction(updateConfigAction)
        tb.addAction(exportDckAction)
        tb.addAction(runSimulationAction)
        tb.addAction(processSimulationAction)
        tb.addAction(deleteDiaAction)

        # Menu bar actions
        self.fileMenu = QMenu("File")

        fileMenuNewAction = QAction("New", self)
        fileMenuNewAction.triggered.connect(self.newDia)
        fileMenuNewAction.setShortcut("Ctrl+n")
        self.fileMenu.addAction(fileMenuNewAction)

        fileMenuOpenAction = QAction("Open", self)
        fileMenuOpenAction.triggered.connect(self.openFile)
        fileMenuOpenAction.setShortcut("Ctrl+o")
        self.fileMenu.addAction(fileMenuOpenAction)

        fileMenuSaveAction = QAction("Save", self)
        fileMenuSaveAction.triggered.connect(self.saveDia)
        fileMenuSaveAction.setShortcut("Ctrl+s")
        self.fileMenu.addAction(fileMenuSaveAction)

        fileMenuCopyToNewAction = QAction("Copy to new folder", self)
        fileMenuCopyToNewAction.triggered.connect(self.copyToNew)
        self.fileMenu.addAction(fileMenuCopyToNewAction)

        exportAsPDF = QAction("Export as PDF", self)
        exportAsPDF.triggered.connect(self.exportPDF)
        exportAsPDF.setShortcut("Ctrl+e")
        self.fileMenu.addAction(exportAsPDF)

        debugConnections = QAction("Debug Conn", self)
        debugConnections.triggered.connect(self.debugConns)
        self.fileMenu.addAction(debugConnections)

        def askForSaveAndLoadSettings() -> None:
            self.askUserForSettingsValuesAndSave()
            self.loadTrnsysPath()

        setTransysPath = QAction("Set TRNSYS path", self)
        setTransysPath.triggered.connect(askForSaveAndLoadSettings)
        self.fileMenu.addAction(setTransysPath)

        self.editMenu = QMenu("Edit")
        self.editMenu.addAction(toggleSnapAction)
        self.editMenu.addAction(toggleAlignModeAction)

        runMassflowSolverActionMenu = QAction("Run mass flow solver", self)
        runMassflowSolverActionMenu.triggered.connect(self.runAndVisMf)

        openVisualizerActionMenu = QAction("Start mass flow visualizer", self)
        openVisualizerActionMenu.triggered.connect(self.visualizeMf)

        exportHydraulicsActionMenu = QAction("Export hydraulic.ddck", self)
        exportHydraulicsActionMenu.triggered.connect(self.exportHydraulicsDdck)

        exportHydCtrlActionMenu = QAction("Export hydraulic_control.ddck", self)
        exportHydCtrlActionMenu.triggered.connect(self.exportHydraulicControl)

        updateConfigActionMenu = QAction("Update run.config", self)
        updateConfigActionMenu.triggered.connect(self.updateRun)

        exportDckActionMenu = QAction("Export dck", self)
        exportDckActionMenu.triggered.connect(self.exportDck)

        runSimulationActionMenu = QAction("Run simulation...", self)
        runSimulationActionMenu.triggered.connect(self.runSimulation)

        processSimulationActionMenu = QAction("Process simulation...", self)
        processSimulationActionMenu.triggered.connect(self.processSimulation)

        exportDdckPlaceHolderValuesJsonFileActionMenu = QAction("Export ddck placeholder values JSON file", self)
        exportDdckPlaceHolderValuesJsonFileActionMenu.triggered.connect(self.exportDdckPlaceHolderValuesJson)

        self.projectMenu = QMenu("Project")
        self.projectMenu.addAction(runMassflowSolverActionMenu)
        self.projectMenu.addAction(openVisualizerActionMenu)
        self.projectMenu.addAction(exportHydraulicsActionMenu)
        self.projectMenu.addAction(exportHydCtrlActionMenu)
        self.projectMenu.addAction(updateConfigActionMenu)
        self.projectMenu.addAction(exportDckActionMenu)
        self.projectMenu.addAction(runSimulationActionMenu)
        self.projectMenu.addAction(processSimulationActionMenu)
        self.projectMenu.addAction(exportDdckPlaceHolderValuesJsonFileActionMenu)

        pytrnsysOnlineDocAction = QAction("pytrnsys online documentation", self)
        pytrnsysOnlineDocAction.triggered.connect(self.openPytrnsysOnlineDoc)

        pytrnsysGuiOnlineDocAction = QAction("GUI online documentation", self)
        pytrnsysGuiOnlineDocAction.triggered.connect(self.openPytrnsysGuiOnlineDoc)

        self.helpMenu = QMenu("Help")
        self.helpMenu.addAction(pytrnsysOnlineDocAction)
        self.helpMenu.addAction(pytrnsysGuiOnlineDocAction)

        # Menu bar
        self.mb = self.menuBar()
        self.mb.addMenu(self.fileMenu)
        self.mb.addMenu(self.projectMenu)
        self.mb.addMenu(self.editMenu)
        self.mb.addMenu(self.helpMenu)
        self.mb.addSeparator()

        # Status bar
        self.sb = self.statusBar()
        self.sb.showMessage("Mode is " + str(self.centralWidget.editorMode))

        # QUndo framework
        self.undoStack = QUndoStack(self)
        undoAction = self.undoStack.createUndoAction(self, "Undo")
        undoAction.setShortcut("Ctrl+z")

        redoAction = self.undoStack.createRedoAction(self, "Redo")
        redoAction.setShortcut("Ctrl+y")

        self.editMenu.addAction(undoAction)
        self.editMenu.addAction(redoAction)

    def newDia(self):
        messageBox = QMessageBox()
        messageBox.setText(
            "Are you sure you want to start a new project? Unsaved progress on the current one will be lost."
        )
        messageBox.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        messageBox.setDefaultButton(QMessageBox.Cancel)

        result = messageBox.exec()
        if result == QMessageBox.Cancel:
            return

        startingDirectoryPath = _pl.Path(self.projectFolder).parent

        createProjectMaybeCancelled = _prj.getCreateProject(startingDirectoryPath)
        if _ccl.isCancelled(createProjectMaybeCancelled):
            return
        createProject = _ccl.value(createProjectMaybeCancelled)

        self.centralWidget = self._createDiagramEditor(createProject)
        self.setCentralWidget(self.centralWidget)

    def saveDia(self):
        self.logger.info("Saving diagram")
        self.centralWidget.save()

    def copyToNew(self):
        currentProjectFolderPath = _pl.Path(self.projectFolder)

        startingDirectoryPath = currentProjectFolderPath.parent

        maybeCancelled = _prj.getExistingEmptyDirectory(startingDirectoryPath)
        if _ccl.isCancelled(maybeCancelled):
            return
        newProjectFolderPath = _ccl.value(maybeCancelled)

        oldProjectFolderPath = _pl.Path(self.projectFolder)

        self._copyContents(oldProjectFolderPath, newProjectFolderPath)
        newJsonFilePath = self._changeAndGetNewJsonFileName(newProjectFolderPath, oldProjectFolderPath)

        loadProject = _prj.LoadProject(newJsonFilePath)

        self.centralWidget = self._createDiagramEditor(loadProject)
        self.setCentralWidget(self.centralWidget)
        self.centralWidget.save(showWarning=False)

    @staticmethod
    def _copyContents(oldProjectFolderPath, newProjectFolderPath):
        for child in oldProjectFolderPath.iterdir():
            destinationPath = newProjectFolderPath / child.name
            if child.is_dir():
                shutil.copytree(child, destinationPath)
            else:
                shutil.copy(child, destinationPath)

    @staticmethod
    def _changeAndGetNewJsonFileName(newProjectFolderPath, oldProjectFolderPath):
        oldJsonFileName = f"{oldProjectFolderPath.name}.json"
        newJsonFileName = f"{newProjectFolderPath.name}.json"
        shutil.move(
            newProjectFolderPath / oldJsonFileName,
            newProjectFolderPath / newJsonFileName,
        )
        newJsonFilePath = newProjectFolderPath / newJsonFileName
        return newJsonFilePath

    def loadDia(self):
        self.logger.info("Loading diagram")
        self.openFile()

    def updateRun(self):
        self.logger.info("Updating run.config")
        configPath = os.path.join(self.projectFolder, "run.config")
        runConfig = configFile(configPath, self.centralWidget)
        runConfig.updateConfig()

    def runSimulation(self):
        ddckPath = os.path.join(self.projectFolder, "ddck")

        #   Check hydraulic.ddck
        hydraulicPath = os.path.join(ddckPath, "hydraulic\\hydraulic.ddck")
        if not os.path.isfile(hydraulicPath):
            self.exportHydraulicsDdck()
        infile = open(hydraulicPath, "r")
        hydraulicLines = infile.readlines()
        blackBoxLines = []
        for i in range(len(hydraulicLines)):
            if "Black box component temperatures" in hydraulicLines[i]:
                j = i + 1
                while hydraulicLines[j] != "\n":
                    blackBoxLines.append(hydraulicLines[j])
                    j += 1
                break
        messageText = "Is this correct?\n\n"
        for line in blackBoxLines:
            messageText += line
        qmb = QMessageBox()
        qmb.setText(messageText)
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        qmb.setDefaultButton(QMessageBox.No)
        ret = qmb.exec()
        if ret == QMessageBox.No:
            qmb = QMessageBox()
            qmb.setText(
                "Please make sure the black box component temperatures are correct in hydraulic.ddck before starting a simluation."
            )
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return

        #   Check ddcks of storage tanks
        storageWithoutFile = []
        for object in self.centralWidget.trnsysObj:
            if isinstance(object, StorageTank):
                storageTankFile = os.path.join(
                    object.displayName, object.displayName + ".ddck"
                )
                storageTankPath = os.path.join(ddckPath, storageTankFile)
                if not (os.path.isfile(storageTankPath)):
                    storageWithoutFile.append(object.displayName + "\n")

        if storageWithoutFile:
            errorMessage = "The following storage tank(s) do(es) not have a corresponding ddck:\n\n"
            for storage in storageWithoutFile:
                errorMessage += storage
            errorMessage += "\nPlease make sure you that you export the ddck for every storage tank before starting a simulation."
            showErrorMessageBox(errorMessage)
            return

        #   Update run.config
        self.updateRun()

        #   Start simulation
        runApp = RunMain()
        executionFailed, errorStatement = runApp.runAction(
            self.logger, self.centralWidget.projectFolder
        )

        if executionFailed:
            errorMessage = (
                f"Exception while trying to execute RunParallelTrnsys:\n\n{errorStatement}"
            )
            showErrorMessageBox(errorMessage)

        return

    def processSimulation(self):
        processPath = os.path.join(self.projectFolder, "process.config")
        if not os.path.isfile(processPath):
            errorMessage = f"No such file: {processPath}"
            showErrorMessageBox(errorMessage)
            return
        processApp = ProcessMain()
        executionFailed, errorStatement = processApp.processAction(
            self.logger, self.centralWidget.projectFolder
        )

        if executionFailed:
            errorMessage = (
                f"Exception while trying to execute RunParallelTrnsys:\n\n{errorStatement}"
            )
            showErrorMessageBox(errorMessage)

        return

    def exportDdckPlaceHolderValuesJson(self):
        result = self.centralWidget.exportDdckPlaceHolderValuesJsonFile()
        if _res.isError(result):
            errorMessage = f"The json file could not be generated: {result.message}"
            showErrorMessageBox(errorMessage)

    def renameDia(self):
        self.logger.info("Renaming diagram...")
        self.centralWidget.showDiagramDlg()

    def deleteDia(self):
        qmb = QMessageBox()
        qmb.setText(
            'Are you sure you want to delete the diagram? (There is no possibility to "undo".)'
        )
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()
        if ret == QMessageBox.Yes:
            self.logger.info("Deleting diagram")
            self.centralWidget.delBlocks()
        else:
            self.logger.info("Canceling")
            return

    def tidyUp(self):
        self.logger.info("Tidying up...")
        self.centralWidget.cleanUpConnections()

    def setZoomIn(self):
        self.logger.info("Setting zoom in")
        self.centralWidget.diagramView.scale(1.2, 1.2)

    def setZoomOut(self):
        self.logger.info("Setting zoom out")
        self.centralWidget.diagramView.scale(0.8, 0.8)

    def setZoom0(self):
        self.logger.info("Setting zoom 0")
        self.centralWidget.diagramView.resetTransform()

    def runAndVisMf(self):
        self.calledByVisualizeMf = True
        filePaths = self.runMassflowSolver()
        if not filePaths:
            self.logger.error("Could not execute runMassflowSolver")
            return

        mfrFile, tempFile = filePaths
        if not os.path.isfile(mfrFile) or not os.path.isfile(tempFile):
            self.logger.error("No mfrFile or tempFile found!")
            return

        MassFlowVisualizer(self, mfrFile, tempFile)
        self.massFlowEnabled = True

    def visualizeMf(self):
        qmb = QMessageBox()
        qmb.setText(
            "Please select the mass flow rate prt-file that you want to visualize."
        )
        qmb.setStandardButtons(QMessageBox.Ok)
        qmb.setDefaultButton(QMessageBox.Ok)
        qmb.exec()

        mfrFile = QFileDialog.getOpenFileName(self, "Open diagram", filter="*Mfr.prt")[
            0
        ].replace("/", "\\")
        tempFile = mfrFile.replace("Mfr", "T")
        self.calledByVisualizeMf = True
        if os.path.isfile(mfrFile) and os.path.isfile(tempFile):
            MassFlowVisualizer(self, mfrFile, tempFile)
            self.massFlowEnabled = True
        else:
            self.logger.info("No mfrFile or tempFile found!")

    def openFile(self):
        self.logger.info("Opening diagram")
        qmb = QMessageBox()
        qmb.setText(
            "Are you sure you want to open another project? Unsaved progress on the current one will be lost."
        )
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()

        if ret == QMessageBox.Cancel:
            return

        maybeCancelled = _prj.getLoadOrMigrateProject()
        if _ccl.isCancelled(maybeCancelled):
            return
        project = _ccl.value(maybeCancelled)

        self.centralWidget = self._createDiagramEditor(project)
        self.setCentralWidget(self.centralWidget)

    def toggleConnLabels(self):
        self.labelVisState = not self.labelVisState
        self.centralWidget.setConnLabelVis(self.labelVisState)

    def exportHydraulicsDdck(self):
        self.centralWidget.exportHydraulics(exportTo="ddck")

    def exportHydraulicControl(self):
        self.centralWidget.exportHydraulicControl()

    def exportDck(self):
        jsonResult = self.centralWidget.exportDdckPlaceHolderValuesJsonFile()
        if _res.isError(jsonResult):
            errorMessage = f"The placeholder values JSON file could not be generated: {jsonResult.message}"
            showErrorMessageBox(errorMessage)
            return

        builder = buildDck.buildDck(self.projectFolder)

        result = builder.buildTrnsysDeck()
        if _res.isError(result):
            errorMessage = f"The deck file could not be generated: {result.message}"
            showErrorMessageBox(errorMessage)

    def toggleEditorMode(self):
        self.logger.info("Toggling editor mode")
        self.centralWidget.editorMode = (self.centralWidget.editorMode + 1) % 2
        self.sb.showMessage("Mode is " + str(self.centralWidget.editorMode))

    def toggleAlignMode(self):
        self.logger.info("Toggling alignMode")
        self.centralWidget.alignMode = not self.centralWidget.alignMode

    def toggleSnap(self):
        self.centralWidget.snapGrid = not self.centralWidget.snapGrid
        self.centralWidget.diagramScene.update()

    def runMassflowSolver(self):
        self.logger.info("Running massflow solver...")

        exportPath = self.centralWidget.exportHydraulics(exportTo="mfs")
        if not exportPath:
            return None

        self.exportedTo = exportPath
        self.logger.info(exportPath)

        if not self.centralWidget.trnsysPath.is_file():
            errorMessage = "TRNExe.exe not found! Consider correcting the path in the settings."
            showErrorMessageBox(errorMessage)
            return None

        try:
            subprocess.run(
                [str(self.centralWidget.trnsysPath), exportPath, "/H"], check=True
            )
            mfrFile = os.path.join(
                self.projectFolder, self.projectFolder.split("\\")[-1] + "_Mfr.prt"
            )
            tempFile = os.path.join(
                self.projectFolder, self.projectFolder.split("\\")[-1] + "_T.prt"
            )
            self.calledByVisualizeMf = False
            return mfrFile, tempFile
        except Exception as exception:
            errorMessage = f"An exception occurred while trying to execute the mass flow solver: {exception}"
            showErrorMessageBox(errorMessage)
            self.logger.error(errorMessage)

            return None

    def movePorts(self):
        self.centralWidget.moveDirectPorts = True

    def mouseMoveEvent(self, e):
        pass

    def openPytrnsysOnlineDoc(self):
        try:
            os.system("start \"\" https://pytrnsys.readthedocs.io")
        except:
            self.logger.warning("Could not open pytrnsys online documentation. (Works on Windows only.)")

    def openPytrnsysGuiOnlineDoc(self):
        try:
            os.system("start \"\" https://spf-ost.github.io/pytrnsys_gui")
        except:
            self.logger.warning("Could not open pytrnsys-gui online documentation. (Works on Windows only.)")

    def exportPDF(self):
        self.centralWidget.printPDF()

    def closeEvent(self, e):
        qmb = QMessageBox()
        qmb.setText(
            "Do you want to save the current state of the project before closing the program?"
        )
        qmb.setStandardButtons(
            QMessageBox.Save | QMessageBox.Close | QMessageBox.Cancel
        )
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()
        if ret == QMessageBox.Cancel:
            e.ignore()
        elif ret == QMessageBox.Close:
            e.accept()
        elif ret == QMessageBox.Save:
            self.centralWidget.save()
            e.accept

    def ensureSettingsExist(self):
        settings = _settings.Settings.tryLoadOrNone()
        if not settings:
            self.askUserForSettingsValuesAndSave()

    def askUserForSettingsValuesAndSave(self):
        newSettings = _sdlg.SettingsDlg.showDialogAndGetSettings(parent=self)
        while newSettings == _sdlg.CANCELLED:
            newSettings = _sdlg.SettingsDlg.showDialogAndGetSettings(parent=self)
        newSettings.save()

    def loadTrnsysPath(self):
        settings = _settings.Settings.load()
        self.centralWidget.trnsysPath = _pl.Path(settings.trnsysBinaryPath)

    def debugConns(self):
        """
        Check each block items for error connections.
        Returns warning message if blockitem contains two input connections or two output connections
        """
        self.logger.info("trnsysObjs:", self.centralWidget.trnsysObj)
        self.noErrorConns = True
        for o in self.centralWidget.trnsysObj:
            self.logger.info(o)
            if isinstance(o, BlockItem) and len(o.outputs) == 1 and len(o.inputs) == 1:
                self.logger.info("Checking block connections", o.displayName)
                objInput = o.inputs[0]
                objOutput = o.outputs[0]
                connToInputToPort = objInput.connectionList[0].toPort
                connToOutputToPort = objOutput.connectionList[0].toPort
                connToInputFromPort = objInput.connectionList[0].fromPort
                connToOutputFromPort = objOutput.connectionList[0].fromPort
                connName1 = objInput.connectionList[0].displayName
                connName2 = objOutput.connectionList[0].displayName
                objName = o.displayName

                if objInput == connToInputToPort and objOutput == connToOutputToPort:
                    msgBox = QMessageBox()
                    msgBox.setText(
                        "both %s and %s are input ports into %s"
                        % (connName1, connName2, objName)
                    )
                    msgBox.exec_()
                    self.noErrorConns = False

                elif (
                        objInput == connToInputFromPort
                        and objOutput == connToOutputFromPort
                ):
                    msgBox = QMessageBox()
                    msgBox.setText(
                        "both %s and %s are output ports from %s"
                        % (connName1, connName2, objName)
                    )
                    msgBox.exec_()
                    self.noErrorConns = False
        return self.noErrorConns

    def _createDiagramEditor(self, project: _prj.Project) -> _de.Editor:
        if isinstance(project, _prj.LoadProject):
            self.projectFolder = str(project.jsonFilePath.parent)
            return _de.Editor(
                self,
                str(project.jsonFilePath.parent),
                jsonPath=None,
                loadValue="load",
                logger=self.logger,
            )

        if isinstance(project, _prj.MigrateProject):
            self.projectFolder = str(project.newProjectFolderPath)
            editor = _de.Editor(
                self,
                str(project.newProjectFolderPath),
                str(project.oldJsonFilePath),
                loadValue="json",
                logger=self.logger,
            )
            editor.save()
            return editor

        if isinstance(project, _prj.CreateProject):
            self.projectFolder = str(project.jsonFilePath.parent)
            return _de.Editor(
                self,
                self.projectFolder,
                jsonPath=str(project.jsonFilePath),
                loadValue="new",
                logger=self.logger,
            )

        raise AssertionError(f"Unknown `project' type: {type(project)}")


def main():
    arguments = args.getArgsOrExit()

    logger = log.setup_custom_logger("root", arguments.logLevel)
    app = QApplication(sys.argv)
    app.setApplicationName("Diagram Creator")

    maybeCancelled = _prj.getProject()
    if _ccl.isCancelled(maybeCancelled):
        return
    project = _ccl.value(maybeCancelled)

    form = _MainWindow(logger, project)
    form.showMaximized()
    form.show()
    form.ensureSettingsExist()
    form.loadTrnsysPath()

    tracer = trc.createTracer(arguments.shallTrace)
    tracer.run(lambda: app.exec())


if __name__ == "__main__":
    main()



import dataclasses as _dc
import typing as _tp

import trnsysGUI.PortItemBase as _pib

import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn
import trnsysGUI.common as _com


@_dc.dataclass
class GlobalNetwork:
    realNodes: _tp.Sequence[_mfn.RealNodeBase]
    internalPortItemToExternalRealNode: _tp.Mapping[_mfn.PortItem, _mfn.RealNodeBase]


def getGlobalNetwork(massFlowContributors: _tp.Sequence[_mfs.MassFlowNetworkContributorMixin]) -> GlobalNetwork:
    allRealNodes, modelsToGraphicalPortItem = _getAllRealNodesAndModelsToGraphicalPortItem(massFlowContributors)

    modelPortItemsToRealNode = {n: rn for rn in allRealNodes for n in rn.getNeighbours() if isinstance(n, _mfn.PortItem)}

    graphicalPortItemsToModels = _getGraphicalPortItemsToModels(modelsToGraphicalPortItem)

    internalPortItemToExternalRealNode = _getInternalPortItemToExternalRealNode(
        allRealNodes, modelsToGraphicalPortItem, graphicalPortItemsToModels, modelPortItemsToRealNode
    )

    return GlobalNetwork(allRealNodes, internalPortItemToExternalRealNode)


def _getAllRealNodesAndModelsToGraphicalPortItem(
    massFlowContributors: _tp.Sequence[_mfs.MassFlowNetworkContributorMixin],
) -> _tp.Tuple[_tp.Sequence[_mfn.RealNodeBase], _tp.Mapping[_mfn.PortItem, _pib.PortItemBase]]:
    allRealNodes: _tp.List[_mfn.RealNodeBase] = []
    modelsToGraphicalPortItem: _tp.Dict[_mfn.PortItem, _pib.PortItemBase] = {}
    for massFlowContributor in massFlowContributors:
        internalPiping = massFlowContributor.getInternalPiping()
        realNodes = internalPiping.getAllRealNodes()
        allRealNodes.extend(realNodes)
        modelsToGraphicalPortItem.update(internalPiping.modelPortItemsToGraphicalPortItem)
    return allRealNodes, modelsToGraphicalPortItem


def _getGraphicalPortItemsToModels(
    modelsToGraphicalPortItem: _tp.Mapping[_mfn.PortItem, _pib.PortItemBase]
) -> _tp.Mapping[_pib.PortItemBase, _tp.Sequence[_mfn.PortItem]]:
    graphicalPortItemsToModels: _tp.Dict[_pib.PortItemBase, _tp.List[_mfn.PortItem]] = {}
    for model, graphicalPortItem in modelsToGraphicalPortItem.items():
        models = graphicalPortItemsToModels.get(graphicalPortItem)

        if models:
            models.append(model)
        else:
            graphicalPortItemsToModels[graphicalPortItem] = [model]
    return graphicalPortItemsToModels


def _getInternalPortItemToExternalRealNode(
    allRealNodes: _tp.Sequence[_mfn.RealNodeBase],
    modelsToGraphicalPortItem: _tp.Mapping[_mfn.PortItem, _pib.PortItemBase],
    graphicalPortItemsToModels: _tp.Mapping[_pib.PortItemBase, _tp.Sequence[_mfn.PortItem]],
    modelPortItemsToRealNode: _tp.Mapping[_mfn.PortItem, _mfn.RealNodeBase],
) -> _tp.Mapping[_mfn.PortItem, _mfn.RealNodeBase]:
    portItemsToJoin = {}
    for internalRealNode in allRealNodes:
        for neighbor in internalRealNode.getNeighbours():
            if not isinstance(neighbor, _mfn.PortItem):
                continue

            portItem = neighbor
            graphicalPortItem = modelsToGraphicalPortItem[portItem]
            candidates = graphicalPortItemsToModels[graphicalPortItem]

            externalPortItem = _com.getSingle(c for c in candidates if c.canOverlapWith(portItem))

            externalRealNode = modelPortItemsToRealNode[externalPortItem]

            portItemsToJoin[portItem] = externalRealNode

    return portItemsToJoin



# pylint: skip-file
# type: ignore

from PyQt5 import QtCore
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QGraphicsEllipseItem

from trnsysGUI.Node import Node


class CornerItem(QGraphicsEllipseItem):
    def __init__(self, x, y, r1, r2, prevNode, nextNode, parent=None):
        """
        CornerItems represent corners for each Connection.
        When a segmentItem is dragged, it will also move the corners, which will trigger the CornerItem
        itemChange method of Qt. There, the adjacent segments are redrawn.

        Note: Using hasattr to detect type of parent of node, better solution needed
        Note: No support for disr (interrupted) segements

        Parameters
        ----------
        x
        y
        r1 : int
        semi-axis length in x direction
        r2 : int
        semi-axis length in y direction
        prevNode
        nextNode
        parent
        """
        super(CornerItem, self).__init__(x, y, r1, r2, None)

        self.logger = parent.logger

        self.logger.debug("init pos is " + str(self.pos()))
        self.parent = parent
        self.setBrush(QBrush(QtCore.Qt.black))
        self.node = Node(self, prevNode, nextNode)

        self.posCallbacks = []

    def itemChange(self, change, value):
        if change == self.ItemScenePositionHasChanged:

            nextNode = self.node.nextN()
            previousNode = self.node.prevN()

            nodePosInConnection = self.parent.getNodePos(self)

            segmentBefore = self.parent.segments[nodePosInConnection]
            segmentAfter = self.parent.segments[nodePosInConnection + 1]

            if type(nextNode.parent) is CornerItem:
                if segmentAfter.line() is not None:
                    segmentAfter.setLine(self.scenePos().x(), self.scenePos().y(), segmentAfter.line().p2().x(),
                                     segmentAfter.line().p2().y())
                else:
                    self.logger.debug("segmentAfter.line() is None")

            if type(previousNode.parent) is CornerItem:
                if segmentBefore.line() is not None:
                    segmentBefore.setLine(segmentBefore.line().p1().x(), segmentBefore.line().p1().y(),
                                          self.scenePos().x(), self.scenePos().y())
                else:
                    self.logger.debug("segmentBefore.line() is None")

            if hasattr(nextNode.parent, "toPort"):
                if nextNode.nextN() is None and self.node.lastNode() is nextNode:
                    self.logger.debug("nextNode is at 'toPort'")
                    lastSegment = self.parent.segments[-1]
                    lastSegment.setLine(self.scenePos().x(), self.scenePos().y(), nextNode.parent.toPort.scenePos().x(),
                                        nextNode.parent.toPort.scenePos().y())

            if hasattr(previousNode.parent, "fromPort"):
                if previousNode.prevN() is None and self.node.firstNode() is previousNode:
                    self.logger.debug("previousNode is at 'fromPort'")
                    firstSegment = self.parent.segments[0]
                    firstSegment.setLine(previousNode.parent.fromPort.scenePos().x(),
                                         previousNode.parent.fromPort.scenePos().y(), self.scenePos().x(),
                                         self.scenePos().y())

            for segment in self.parent.segments:
                try:
                    segment.updateGrad()
                except:
                    self.logger.warning('Could not update color gradient of pipe.')

            for callback in self.posCallbacks:
                callback(value)
            return value

        return super(CornerItem, self).itemChange(change, value)


import typing as _tp

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.BlockItem as _bi
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn

import trnsysGUI.connectorsAndPipesExportHelpers as _helpers


class Connector(_bi.BlockItem, _mfs.MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self._updateModelPipe(self.displayName)

        self.sizeFactor = 0.5
        self.w = 40
        self.h = 40

        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))

        self.changeSize()

    def getInternalPiping(self) -> _mfs.InternalPiping:
        return _mfs.InternalPiping(
            [self._modelPipe], {self._modelPipe.fromPort: self.inputs[0], self._modelPipe.toPort: self.outputs[0]}
        )

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.CONNECTOR_PNG

    def setDisplayName(self, newName: str) -> None:
        super().setDisplayName(newName)
        self._updateModelPipe(newName)

    def changeSize(self):
        width = self.w
        height = self.h

        delta = 20  # pylint: disable=duplicate-code  # 1
        # Limit the block size:
        height = max(height, 20)
        width = max(width, 40)

        # center label:
        rect = self.label.boundingRect()
        labelWidth = rect.width()  # pylint: disable=duplicate-code  # 1
        labelPosX = (width - labelWidth) / 2
        self.label.setPos(labelPosX, height)

        self.origInputsPos = [[0, delta]]
        self.origOutputsPos = [[width, delta]]  # pylint: disable=duplicate-code  # 1
        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4  # pylint: disable=duplicate-code  # 1

        return width, height

    def exportBlackBox(self):
        return "noBlackBoxOutput", []

    def exportPipeAndTeeTypesForTemp(self, startingUnit: int) -> _tp.Tuple[str, int]:
        tempName = self.getTemperatureVariableName(self.inputs[0])
        mfrName = _helpers.getMfrName(self._modelPipe)

        fromConnection = self.inputs[0].getConnection()
        toConnection = self.outputs[0].getConnection()

        posFlowTempName = f"T{fromConnection.displayName}"
        negFlowTempName = f"T{toConnection.displayname}"
        equation = f"{tempName} = GE({mfrName}, 0)*{posFlowTempName} + LT({mfrName}, 0)*{negFlowTempName}"
        equations = f"""\
EQUATIONS 1
{equation}
"""
        return equations, startingUnit

    def _updateModelPipe(self, displayName: str) -> None:
        fromPort = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        toPort = _mfn.PortItem("output", _mfn.PortItemType.OUTPUT)
        self._modelPipe = _mfn.Pipe(displayName, self.trnsysId, fromPort, toPort)



import itertools as _it


class IdGenerator:  # pylint: disable=too-many-instance-attributes
    # This value is used for the IDs when migrating from old JSON formats.
    # After deserialization the IDs are set such that there are no duplicate
    # IDs across all objects.
    UNINITIALIZED_ID = -1

    def __init__(self):
        self.ids = []
        self.trnsysIds = []
        self.connIds = []
        self.storageTes = []
        self.storageType = []
        self.idCounter = _it.count(1)
        self.trnsysIdCounter = _it.count(1)
        self.connIdCounter = _it.count(1)
        self.storagenTesCounter = _it.count(1)
        self.storageTypeCounter = _it.count(1924)

    def getID(self):
        nextId = next(self.idCounter)
        self.ids.append(nextId)
        return nextId

    def getTrnsysID(self):
        nextId = next(self.trnsysIdCounter)
        self.trnsysIds.append(nextId)
        return nextId

    def getConnID(self):
        nextId = next(self.connIdCounter)
        self.connIds.append(nextId)
        return nextId

    def getStoragenTes(self):
        nextId = next(self.storagenTesCounter)
        self.storageTes.append(nextId)
        return nextId

    def getStorageType(self):
        nextId = next(self.storageTypeCounter)
        self.storageType.append(nextId)
        return nextId

    def setID(self, nextId):
        self.idCounter = _it.count(nextId)

    def setConnID(self, nextId):
        self.connIdCounter = _it.count(nextId)

    def setTrnsysID(self, nextId):
        self.trnsysIdCounter = _it.count(nextId)

    def reset(self):
        self.ids = []
        self.trnsysIds = []
        self.connIds = []
        self.storageTes = []
        self.storageType = []
        self.idCounter = _it.count(1)
        self.trnsysIdCounter = _it.count(1)
        self.connIdCounter = _it.count(1)
        self.storagenTesCounter = _it.count(1)
        self.storageTypeCounter = _it.count(1924)



from __future__ import annotations

import typing as _tp

import PyQt5.QtWidgets as _qtw

import pytrnsys.utils.result as _res

import trnsysGUI.errors as _err
import trnsysGUI.connection.singlePipeConnection as _spc
import trnsysGUI.hydraulicLoops.merge as _hlmerge
import trnsysGUI.hydraulicLoops.split as _hlsplit
import trnsysGUI.singlePipePortItem as _spi

if _tp.TYPE_CHECKING:
    import trnsysGUI.diagram.Editor as _ed


class CreateSinglePipeConnectionCommand(_qtw.QUndoCommand):
    def __init__(
            self,
            fromPort: _spi.SinglePipePortItem,
            toPort: _spi.SinglePipePortItem,
            editor: _ed.Editor,  # type: ignore[name-defined]
    ):
        super().__init__("Create single pipe connection")
        self._fromPort = fromPort
        self._toPort = toPort
        self._editor = editor

        self._connection: _tp.Optional[_spc.SinglePipeConnection] = None
        self._mergeSummary: _tp.Optional[_hlmerge.MergeSummary] = None

    def redo(self):
        self._connection = _spc.SinglePipeConnection(self._fromPort, self._toPort, self._editor)

        mergedLoopSummary = self._mergeSummary.after if self._mergeSummary else None  # pylint: disable=no-member

        cancellable = _hlmerge.merge(
            self._connection,
            self._editor.hydraulicLoops,
            self._editor.fluids.fluids,
            self._editor.fluids.WATER,
            mergedLoopSummary,
        )
        if cancellable == "cancelled" or _res.isError(cancellable):
            if _res.isError(cancellable):
                error = _res.error(cancellable)
                _err.showErrorMessageBox(error.message, "Cannot create connection")

            self._connection.deleteConn()
            self._connection = None
            self.setObsolete(True)
            return

        mergeSummary = cancellable

        self._mergeSummary = mergeSummary

    def undo(self):
        splitLoopsSummary = self._mergeSummary.before if self._mergeSummary else None  # pylint: disable=no-member

        cancellable = _hlsplit.split(
            self._connection, self._editor.hydraulicLoops, self._editor.fluids.fluids, splitLoopsSummary
        )
        assert cancellable != "cancelled"

        self._connection.deleteConn()
        self._connection = None



from trnsysGUI import singlePipePortItem as _spi


def getFromAndToPort(connection):
    fromPort = connection.fromPort
    toPort = connection.toPort
    assert isinstance(fromPort, _spi.SinglePipePortItem)
    assert isinstance(toPort, _spi.SinglePipePortItem)
    return fromPort, toPort



import typing as _tp
import pathlib as _pl

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem  # type: ignore[attr-defined]
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class SourceSinkBase(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.w = 60
        self.h = 60

        self.inputs.append(_cspi.createSinglePipePortItem("i", 1, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 1, self))

        self.changeSize()

        ddckFilePath = _pl.Path(self.parent.parent().projectPath) / "ddck" / f"{self.displayName}.ddck"

        self.path = str(ddckFilePath)

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        raise NotImplementedError()

    def changeSize(self):
        self._positionLabel()

        self.origInputsPos = [[20, 0]]
        self.origOutputsPos = [[40, 0]]

        # pylint: disable=duplicate-code  # 1
        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 1 + 2 * self.flippedV) % 4
        self.outputs[0].side = (self.rotationN + 1 + 2 * self.flippedV) % 4
        # pylint: disable=duplicate-code  # 1

    def exportBlackBox(self):
        status = "noDdckEntry"
        equation = ["T" + self.displayName + "=1"]

        return status, equation

    def getInternalPiping(self) -> InternalPiping:
        inputPort = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        outputPort = _mfn.PortItem("output", _mfn.PortItemType.OUTPUT)

        pump = _mfn.Pump(self.displayName, self.trnsysId, inputPort, outputPort)

        modelPortItemsToGraphicalPortItem = {inputPort: self.inputs[0], outputPort: self.outputs[0]}
        return InternalPiping([pump], modelPortItemsToGraphicalPortItem)

    def exportMassFlows(self):
        equationNr = 1
        massFlowLine = f"Mfr{self.displayName} = 500\n"
        return massFlowLine, equationNr

    def hasDdckPlaceHolders(self):
        return True



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'trnsysGUI/hydraulicLoops/_dialogs/split/_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.3
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_splitHydraulicLoopDialog(object):
    def setupUi(self, splitHydraulicLoopDialog):
        splitHydraulicLoopDialog.setObjectName("splitHydraulicLoopDialog")
        splitHydraulicLoopDialog.resize(353, 209)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(splitHydraulicLoopDialog.sizePolicy().hasHeightForWidth())
        splitHydraulicLoopDialog.setSizePolicy(sizePolicy)
        splitHydraulicLoopDialog.setMaximumSize(QtCore.QSize(16777215, 209))
        self.verticalLayout = QtWidgets.QVBoxLayout(splitHydraulicLoopDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.widget = QtWidgets.QWidget(splitHydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setMinimumSize(QtCore.QSize(20, 20))
        self.widget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-information.svg);")
        self.widget.setObjectName("widget")
        self.horizontalLayout_3.addWidget(self.widget)
        self.label = QtWidgets.QLabel(splitHydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.loop1GroupBox = EnterLeaveNotifyingGroupBox(splitHydraulicLoopDialog)
        self.loop1GroupBox.setObjectName("loop1GroupBox")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.loop1GroupBox)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.loopNameLabel = QtWidgets.QLabel(self.loop1GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loopNameLabel.sizePolicy().hasHeightForWidth())
        self.loopNameLabel.setSizePolicy(sizePolicy)
        self.loopNameLabel.setObjectName("loopNameLabel")
        self.horizontalLayout.addWidget(self.loopNameLabel)
        self.name1ComboBox = QtWidgets.QComboBox(self.loop1GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.name1ComboBox.sizePolicy().hasHeightForWidth())
        self.name1ComboBox.setSizePolicy(sizePolicy)
        self.name1ComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.name1ComboBox.setEditable(True)
        self.name1ComboBox.setObjectName("name1ComboBox")
        self.horizontalLayout.addWidget(self.name1ComboBox)
        self.name1WarningWidget = QtWidgets.QWidget(self.loop1GroupBox)
        self.name1WarningWidget.setMinimumSize(QtCore.QSize(30, 30))
        self.name1WarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.name1WarningWidget.setObjectName("name1WarningWidget")
        self.horizontalLayout.addWidget(self.name1WarningWidget)
        spacerItem = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.fluidLabel = QtWidgets.QLabel(self.loop1GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidLabel.sizePolicy().hasHeightForWidth())
        self.fluidLabel.setSizePolicy(sizePolicy)
        self.fluidLabel.setObjectName("fluidLabel")
        self.horizontalLayout.addWidget(self.fluidLabel)
        self.fluid1ComboBox = QtWidgets.QComboBox(self.loop1GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluid1ComboBox.sizePolicy().hasHeightForWidth())
        self.fluid1ComboBox.setSizePolicy(sizePolicy)
        self.fluid1ComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.fluid1ComboBox.setObjectName("fluid1ComboBox")
        self.horizontalLayout.addWidget(self.fluid1ComboBox)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addWidget(self.loop1GroupBox)
        self.loop2GroupBox = EnterLeaveNotifyingGroupBox(splitHydraulicLoopDialog)
        self.loop2GroupBox.setObjectName("loop2GroupBox")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.loop2GroupBox)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.loopNameLabel_2 = QtWidgets.QLabel(self.loop2GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loopNameLabel_2.sizePolicy().hasHeightForWidth())
        self.loopNameLabel_2.setSizePolicy(sizePolicy)
        self.loopNameLabel_2.setObjectName("loopNameLabel_2")
        self.horizontalLayout_2.addWidget(self.loopNameLabel_2)
        self.name2ComboBox = QtWidgets.QComboBox(self.loop2GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.name2ComboBox.sizePolicy().hasHeightForWidth())
        self.name2ComboBox.setSizePolicy(sizePolicy)
        self.name2ComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.name2ComboBox.setEditable(True)
        self.name2ComboBox.setObjectName("name2ComboBox")
        self.horizontalLayout_2.addWidget(self.name2ComboBox)
        self.name2WarningWidget = QtWidgets.QWidget(self.loop2GroupBox)
        self.name2WarningWidget.setMinimumSize(QtCore.QSize(30, 30))
        self.name2WarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.name2WarningWidget.setObjectName("name2WarningWidget")
        self.horizontalLayout_2.addWidget(self.name2WarningWidget)
        spacerItem2 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem2)
        self.fluidLabel_2 = QtWidgets.QLabel(self.loop2GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidLabel_2.sizePolicy().hasHeightForWidth())
        self.fluidLabel_2.setSizePolicy(sizePolicy)
        self.fluidLabel_2.setObjectName("fluidLabel_2")
        self.horizontalLayout_2.addWidget(self.fluidLabel_2)
        self.fluid2ComboBox = QtWidgets.QComboBox(self.loop2GroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluid2ComboBox.sizePolicy().hasHeightForWidth())
        self.fluid2ComboBox.setSizePolicy(sizePolicy)
        self.fluid2ComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.fluid2ComboBox.setObjectName("fluid2ComboBox")
        self.horizontalLayout_2.addWidget(self.fluid2ComboBox)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem3)
        self.verticalLayout.addWidget(self.loop2GroupBox)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem4)
        self.applyButton = QtWidgets.QPushButton(splitHydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.applyButton.sizePolicy().hasHeightForWidth())
        self.applyButton.setSizePolicy(sizePolicy)
        self.applyButton.setObjectName("applyButton")
        self.horizontalLayout_5.addWidget(self.applyButton)
        self.abortButton = QtWidgets.QPushButton(splitHydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.abortButton.sizePolicy().hasHeightForWidth())
        self.abortButton.setSizePolicy(sizePolicy)
        self.abortButton.setObjectName("abortButton")
        self.horizontalLayout_5.addWidget(self.abortButton)
        self.verticalLayout.addLayout(self.horizontalLayout_5)

        self.retranslateUi(splitHydraulicLoopDialog)
        QtCore.QMetaObject.connectSlotsByName(splitHydraulicLoopDialog)

    def retranslateUi(self, splitHydraulicLoopDialog):
        _translate = QtCore.QCoreApplication.translate
        splitHydraulicLoopDialog.setWindowTitle(_translate("splitHydraulicLoopDialog", "Split hydraulic loop"))
        self.label.setText(_translate("splitHydraulicLoopDialog", "Hover over the designated area of a given loop to highlight its connections."))
        self.loop1GroupBox.setTitle(_translate("splitHydraulicLoopDialog", "Loop 1"))
        self.loopNameLabel.setText(_translate("splitHydraulicLoopDialog", "Name"))
        self.fluidLabel.setText(_translate("splitHydraulicLoopDialog", "Fluid"))
        self.loop2GroupBox.setTitle(_translate("splitHydraulicLoopDialog", "Loop 2"))
        self.loopNameLabel_2.setText(_translate("splitHydraulicLoopDialog", "Name"))
        self.fluidLabel_2.setText(_translate("splitHydraulicLoopDialog", "Fluid"))
        self.applyButton.setText(_translate("splitHydraulicLoopDialog", "Apply"))
        self.abortButton.setText(_translate("splitHydraulicLoopDialog", "Abort"))
from ._widgets import EnterLeaveNotifyingGroupBox



# pylint: skip-file
# type: ignore

import json
import math as _math
import os
import pathlib as _pl
import pkgutil as _pu
import shutil
import typing as _tp

from PyQt5 import QtGui
from PyQt5.QtCore import QSize, Qt, QLineF, QFileInfo, QDir, QPointF, QEvent
from PyQt5.QtGui import QColor, QPainter
from PyQt5.QtPrintSupport import QPrinter
from PyQt5.QtSvg import QSvgGenerator
from PyQt5.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QListView,
    QVBoxLayout,
    QListWidget,
    QLabel,
    QLineEdit,
    QScrollArea,
    QSplitter,
    QGraphicsItemGroup,
    QGraphicsLineItem,
    QMessageBox,
    QFileDialog,
    QTreeView,
    QPushButton,
)

import pytrnsys.trnsys_util.deckUtils as _du
import pytrnsys.utils.result as _res
import trnsysGUI as _tgui
import trnsysGUI.diagram.Encoder as _enc
import trnsysGUI.errors as _errs
import trnsysGUI.hydraulicLoops.edit as _hledit
import trnsysGUI.hydraulicLoops.migration as _hlmig
import trnsysGUI.hydraulicLoops.model as _hlm
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.placeholders as _ph
from trnsysGUI.BlockDlg import BlockDlg
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.Export import Export
from trnsysGUI.FileOrderingDialog import FileOrderingDialog
from trnsysGUI.GenericPortPairDlg import GenericPortPairDlg
from trnsysGUI.Graphicaltem import GraphicalItem
from trnsysGUI.LibraryModel import LibraryModel
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel
from trnsysGUI.MyQTreeView import MyQTreeView
from trnsysGUI.PortItemBase import PortItemBase
from trnsysGUI.PumpDlg import PumpDlg
from trnsysGUI.TVentil import TVentil
from trnsysGUI.TVentilDlg import TVentilDlg
from trnsysGUI.connection.connectionBase import ConnectionBase
from trnsysGUI.connection.createDoublePipeConnectionCommand import CreateDoublePipeConnectionCommand
from trnsysGUI.connection.createSinglePipeConnectionCommand import CreateSinglePipeConnectionCommand
from trnsysGUI.connection.singlePipeConnection import SinglePipeConnection
from trnsysGUI.diagram.Decoder import Decoder
from trnsysGUI.diagram.scene import Scene
from trnsysGUI.diagram.view import View
from trnsysGUI.diagramDlg import diagramDlg
from trnsysGUI.doublePipeBlockDlg import DoublePipeBlockDlg
from trnsysGUI.doublePipePortItem import DoublePipePortItem
from trnsysGUI.hxDlg import hxDlg
from trnsysGUI.idGenerator import IdGenerator
from trnsysGUI.segmentDlg import segmentDlg
from trnsysGUI.singlePipePortItem import SinglePipePortItem
from trnsysGUI.storageTank.ConfigureStorageDialog import ConfigureStorageDialog
from trnsysGUI.storageTank.widget import StorageTank


class Editor(QWidget):
    """
    This class is the central widget of the MainWindow.
    It contains the items library, diagram graphics scene and graphics view, and the inspector widget

    Function of Connections:
    Logically:
    A Connection is composed of a fromPort and a toPort, which gives the direction of the pipe.
    Ports are attached to Blocks.
    Visually:
    A diagram editor has a QGraphicsLineItem (connLineItem) which is set Visible only when a connection is being created

    Function of BlockItems:
    Items can be added to the library by adding them to the model of the library broswer view.
    Then they can be dragged and dropped into the diagram view.

    Function of trnsysExport:
    When exporting the trnsys file, exportData() is called.

    Function of save and load:
    A diagram can be saved to a json file by calling encodeDiagram and can then be loaded by calling decodeDiagram wiht
    appropiate filenames.

    Attributes
    ----------
    projectFolder : str
        Path to the folder of the project
    diagramName : str
        Name used for saving the diagram
    saveAsPath : :obj:`Path`
        Default saving location is trnsysGUI/diagrams, path only set if "save as" used
    idGen : :obj:`IdGenerator`
        Is used to distribute ids (id, trnsysId(for trnsysExport), etc)
    alignMode : bool
        Enables mode in which a dragged block is aligned to y or x value of another one
        Toggled in the MainWindow class in toggleAlignMode()

    editorMode : int
        Mode 0: Pipes are PolySun-like
        Mode 1: Pipes have only 90deg angles, visio-like
    snapGrid : bool
        Enable/Disable align grid
    snapSize : int
        Size of align grid

    horizontalLayout : :obj:`QHBoxLayout`
    Contains the diagram editor and the layout containing the library browser view and the listview
    vertL : :obj:`QVBoxLayout`
    Cointains the library browser view and the listWidget

    moveDirectPorts: bool
        Enables/Disables moving direct ports of storagetank (doesn't work with HxPorts yet)
    diagramScene : :obj:`QGraphicsScene`
        Contains the "logical" part of the diagram
    diagramView : :obj:`QGraphicsView`
        Contains the visualization of the diagramScene
    _currentlyDraggedConnectionFromPort : :obj:`PortItem`
    connectionList : :obj:`List` of :obj:`Connection`
    trnsysObj : :obj:`List` of :obj:`BlockItem` and :obj:`Connection`
    graphicalObj : :obj:`List` of :obj:`GraphicalItem`
    connLine : :obj:`QLineF`
    connLineItem = :obj:`QGraphicsLineItem`

    """

    def __init__(self, parent, projectFolder, jsonPath, loadValue, logger):
        super().__init__(parent)

        self.logger = logger

        self.logger.info("Initializing the diagram editor")

        self.projectFolder = projectFolder

        self.diagramName = os.path.split(self.projectFolder)[-1] + ".json"
        self.saveAsPath = _pl.Path()
        self.idGen = IdGenerator()

        self.testEnabled = False
        self.existReference = True

        self.controlExists = 0
        self.controlDirectory = ""

        self.alignMode = False

        self.moveDirectPorts = False

        self.editorMode = 1

        # Related to the grid blocks can snap to
        self.snapGrid = False
        self.snapSize = 20

        self.trnsysPath = _pl.Path(r"C:\Trnsys18\Exe\TRNExe.exe")

        self.horizontalLayout = QHBoxLayout(self)
        self.libraryBrowserView = QListView(self)
        self.libraryModel = LibraryModel(self)

        self.libraryBrowserView.setGridSize(QSize(65, 65))
        self.libraryBrowserView.setResizeMode(QListView.Adjust)
        self.libraryModel.setColumnCount(0)

        componentNamesWithIcon = [
            ("Connector", _img.CONNECTOR_SVG.icon()),
            ("TeePiece", _img.TEE_PIECE_SVG.icon()),
            ("DPTee", _img.DP_TEE_PIECE_SVG.icon()),
            ("SPCnr", _img.SINGLE_DOUBLE_PIPE_CONNECTOR_SVG.icon()),
            ("DPCnr", _img.DOUBLE_DOUBLE_PIPE_CONNECTOR_SVG.icon()),
            ("TVentil", _img.T_VENTIL_SVG.icon()),
            ("WTap_main", _img.W_TAP_MAIN_SVG.icon()),
            ("WTap", _img.W_TAP_SVG.icon()),
            ("Pump", _img.PUMP_SVG.icon()),
            ("Collector", _img.COLLECTOR_SVG.icon()),
            ("GroundSourceHx", _img.GROUND_SOURCE_HX_SVG.icon()),
            ("PV", _img.PV_SVG.icon()),
            ("HP", _img.HP_SVG.icon()),
            ("HPTwoHx", _img.HP_TWO_HX_SVG.icon()),
            ("HPDoubleDual", _img.HP_DOUBLE_DUAL_SVG.icon()),
            ("HPDual", _img.HP_DUAL_SVG.icon()),
            ("AirSourceHP", _img.AIR_SOURCE_HP_SVG.icon()),
            ("StorageTank", _img.STORAGE_TANK_SVG.icon()),
            ("IceStorage", _img.ICE_STORAGE_SVG.icon()),
            ("PitStorage", _img.PIT_STORAGE_SVG.icon()),
            ("IceStorageTwoHx", _img.ICE_STORAGE_TWO_HX_SVG.icon()),
            ("ExternalHx", _img.EXTERNAL_HX_SVG.icon()),
            ("Radiator", _img.RADIATOR_SVG.icon()),
            ("Boiler", _img.BOILER_SVG.icon()),
            ("Sink", _img.SINK_SVG.icon()),
            ("Source", _img.SOURCE_SVG.icon()),
            ("SourceSink", _img.SOURCE_SINK_SVG.icon()),
            ("Geotherm", _img.GEOTHERM_SVG.icon()),
            ("Water", _img.WATER_SVG.icon()),
            ("Crystalizer", _img.CRYSTALIZER_SVG.icon()),
            ("GenericBlock", _img.GENERIC_BLOCK_PNG.icon()),
            ("GraphicalItem", _img.GENERIC_ITEM_PNG.icon()),
        ]

        libItems = [QtGui.QStandardItem(icon, name) for name, icon in componentNamesWithIcon]

        for i in libItems:
            self.libraryModel.appendRow(i)

        self.libraryBrowserView.setModel(self.libraryModel)
        self.libraryBrowserView.setViewMode(self.libraryBrowserView.IconMode)
        self.libraryBrowserView.setDragDropMode(self.libraryBrowserView.DragOnly)

        self.diagramScene = Scene(self)
        self.diagramView = View(self.diagramScene, self)

        # For list view
        self.vertL = QVBoxLayout()
        self.vertL.addWidget(self.libraryBrowserView)
        self.vertL.setStretchFactor(self.libraryBrowserView, 2)
        self.listV = QListWidget()
        self.vertL.addWidget(self.listV)
        self.vertL.setStretchFactor(self.listV, 1)

        # for file browser
        self.projectPath = ""
        self.fileList = []

        if loadValue == "new" or loadValue == "json":
            self.createProjectFolder()

        self.fileBrowserLayout = QVBoxLayout()
        self.pathLayout = QHBoxLayout()
        self.projectPathLabel = QLabel("Project Path:")
        self.PPL = QLineEdit(self.projectFolder)
        self.PPL.setDisabled(True)

        self.pathLayout.addWidget(self.projectPathLabel)
        self.pathLayout.addWidget(self.PPL)
        self.scroll = QScrollArea()
        self.scroll.setWidgetResizable(True)
        self.splitter = QSplitter(
            Qt.Vertical,
        )
        self.splitter.setChildrenCollapsible(False)
        self.scroll.setWidget(self.splitter)
        self.scroll.setFixedWidth(350)
        self.fileBrowserLayout.addLayout(self.pathLayout)
        self.fileBrowserLayout.addWidget(self.scroll)
        self.createDdckTree(self.projectFolder)

        if loadValue == "new" or loadValue == "json":
            self.createConfigBrowser(self.projectFolder)
            self.copyGenericFolder(self.projectFolder)
            self.createHydraulicDir(self.projectFolder)
            self.createWeatherAndControlDirs(self.projectFolder)

        self.horizontalLayout.addLayout(self.vertL)
        self.horizontalLayout.addWidget(self.diagramView)
        self.horizontalLayout.addLayout(self.fileBrowserLayout)
        self.horizontalLayout.setStretchFactor(self.diagramView, 5)
        self.horizontalLayout.setStretchFactor(self.libraryBrowserView, 1)

        self._currentlyDraggedConnectionFromPort = None
        self.connectionList = []
        self.trnsysObj = []
        self.graphicalObj = []
        self.fluids = _hlm.Fluids([])
        self.hydraulicLoops = _hlm.HydraulicLoops([])

        self.copyGroupList = QGraphicsItemGroup()
        self.selectionGroupList = QGraphicsItemGroup()

        self.printerUnitnr = 0

        # Different colors for connLineColor
        colorsc = "red"
        linePx = 4
        if colorsc == "red":
            connLinecolor = QColor(Qt.red)
        elif colorsc == "blueish":
            connLinecolor = QColor(3, 124, 193)  # Blue
        elif colorsc == "darkgray":
            connLinecolor = QColor(140, 140, 140)  # Gray
        else:
            connLinecolor = QColor(196, 196, 196)  # Gray

        # Only for displaying on-going creation of connection
        self.connLine = QLineF()
        self.connLineItem = QGraphicsLineItem(self.connLine)
        self.connLineItem.setPen(QtGui.QPen(connLinecolor, linePx))
        self.connLineItem.setVisible(False)
        self.diagramScene.addItem(self.connLineItem)

        # For line that shows quickly up when using the align mode
        self.alignYLine = QLineF()
        self.alignYLineItem = QGraphicsLineItem(self.alignYLine)
        self.alignYLineItem.setPen(QtGui.QPen(QColor(196, 249, 252), 2))
        self.alignYLineItem.setVisible(False)
        self.diagramScene.addItem(self.alignYLineItem)

        # For line that shows quickly up when using align mode
        self.alignXLine = QLineF()
        self.alignXLineItem = QGraphicsLineItem(self.alignXLine)
        self.alignXLineItem.setPen(QtGui.QPen(QColor(196, 249, 252), 2))
        self.alignXLineItem.setVisible(False)
        self.diagramScene.addItem(self.alignXLineItem)

        if loadValue == "load" or loadValue == "copy":
            self._decodeDiagram(os.path.join(self.projectFolder, self.diagramName), loadValue=loadValue)
        elif loadValue == "json":
            self._decodeDiagram(jsonPath, loadValue=loadValue)

    # Debug function
    def dumpInformation(self):
        self.logger.debug("Diagram information:")
        self.logger.debug("Mode is " + str(self.editorMode))

        self.logger.debug("Next ID is " + str(self.idGen.getID()))
        self.logger.debug("Next cID is " + str(self.idGen.getConnID()))

        self.logger.debug("TrnsysObjects are:")
        for t in self.trnsysObj:
            self.logger.debug(str(t))
        self.logger.debug("")

        self.logger.debug("Scene items are:")
        sItems = self.diagramScene.items()
        for it in sItems:
            self.logger.info(str(it))
        self.logger.debug("")

        for c in self.connectionList:
            c.printConn()
        self.logger.debug("")

    # Connections related methods
    def startConnection(self, port):
        self._currentlyDraggedConnectionFromPort = port

    def _createConnection(self, startPort, endPort) -> None:
        if startPort is not endPort:
            if (
                isinstance(startPort.parent, StorageTank)
                and isinstance(endPort.parent, StorageTank)
                and startPort.parent != endPort.parent
            ):
                msgSTank = QMessageBox(self)
                msgSTank.setText("Storage Tank to Storage Tank connection is not working atm!")
                msgSTank.exec_()

            isValidSinglePipeConnection = isinstance(startPort, SinglePipePortItem) and isinstance(
                endPort, SinglePipePortItem
            )
            if isValidSinglePipeConnection:
                command = CreateSinglePipeConnectionCommand(startPort, endPort, self)
            elif isinstance(startPort, DoublePipePortItem) and isinstance(endPort, DoublePipePortItem):
                command = CreateDoublePipeConnectionCommand(startPort, endPort, self)
            else:
                raise AssertionError("Can only connect port items. Also, they have to be of the same type.")

            self.parent().undoStack.push(command)

    def sceneMouseMoveEvent(self, event):
        """
        This function is for dragging and connecting one port to another.
        When dragging, the fromPort will remain enlarged and black in color and when the toPort is hovered over, it will be
        enlarged and turn red.
        A port's details will also be displayed at the widget when they are hovered over.
        """
        fromPort = self._currentlyDraggedConnectionFromPort
        if not fromPort:
            return

        fromX = fromPort.scenePos().x()
        fromY = fromPort.scenePos().y()

        toX = event.scenePos().x()
        toY = event.scenePos().y()

        self.connLine.setLine(fromX, fromY, toX, toY)
        self.connLineItem.setLine(self.connLine)
        self.connLineItem.setVisible(True)

        hitPortItem = self._getHitPortItemOrNone(event)

        if not hitPortItem:
            return

        mousePosition = event.scenePos()

        portItemX = hitPortItem.scenePos().x()
        portItemY = hitPortItem.scenePos().y()

        distance = _math.sqrt((mousePosition.x() - portItemX) ** 2 + (mousePosition.y() - portItemY) ** 2)
        if distance <= 3.5:
            hitPortItem.enlargePortSize()
            hitPortItem.innerCircle.setBrush(hitPortItem.ashColorR)
            self.listV.clear()
            hitPortItem.debugprint()
        else:
            hitPortItem.resetPortSize()
            hitPortItem.innerCircle.setBrush(hitPortItem.visibleColor)
            self.listV.clear()
            fromPort.debugprint()

        fromPort.enlargePortSize()
        fromPort.innerCircle.setBrush(hitPortItem.visibleColor)

    def _getHitPortItemOrNone(self, event: QEvent) -> _tp.Optional[PortItemBase]:
        fromPort = self._currentlyDraggedConnectionFromPort
        mousePosition = event.scenePos()

        relevantPortItems = self._getRelevantHitPortItems(mousePosition, fromPort)
        if not relevantPortItems:
            return None

        numberOfHitPortsItems = len(relevantPortItems)
        if numberOfHitPortsItems > 1:
            raise NotImplementedError("Can't deal with overlapping port items.")

        hitPortItem = relevantPortItems[0]

        return hitPortItem

    def sceneMouseReleaseEvent(self, event):
        if not self._currentlyDraggedConnectionFromPort:
            return
        fromPort = self._currentlyDraggedConnectionFromPort

        self._currentlyDraggedConnectionFromPort = None
        self.connLineItem.setVisible(False)

        mousePosition = event.scenePos()
        relevantPortItems = self._getRelevantHitPortItems(mousePosition, fromPort)

        numberOfHitPortsItems = len(relevantPortItems)

        if numberOfHitPortsItems > 1:
            raise NotImplementedError("Can't deal with overlapping port items.")

        if numberOfHitPortsItems == 1:
            toPort = relevantPortItems[0]

            if toPort != fromPort:
                self._createConnection(fromPort, toPort)

    def _getRelevantHitPortItems(self, mousePosition: QPointF, fromPort: PortItemBase) -> _tp.Sequence[PortItemBase]:
        hitItems = self.diagramScene.items(mousePosition)
        relevantPortItems = [
            i for i in hitItems if isinstance(i, PortItemBase) and type(i) == type(fromPort) and not i.connectionList
        ]
        return relevantPortItems

    def cleanUpConnections(self):
        for c in self.connectionList:
            c.niceConn()

    def exportHydraulics(self, exportTo=_tp.Literal["ddck", "mfs"]):
        assert exportTo in ["ddck", "mfs"]

        if not self._isHydraulicConnected():
            messageBox = QMessageBox()
            messageBox.setWindowTitle("Hydraulic not connected")
            messageBox.setText("You need to connect all port items before you can export the hydraulics.")
            messageBox.setStandardButtons(QMessageBox.Ok)
            messageBox.exec()
            return

        self.logger.info("------------------------> START OF EXPORT <------------------------")

        self.sortTrnsysObj()

        fullExportText = ""

        ddckFolder = os.path.join(self.projectFolder, "ddck")

        if exportTo == "mfs":
            mfsFileName = self.diagramName.rsplit(".", 1)[0] + "_mfs.dck"
            exportPath = os.path.join(self.projectFolder, mfsFileName)
        elif exportTo == "ddck":
            exportPath = os.path.join(ddckFolder, "hydraulic", "hydraulic.ddck")

        if self._doesFileExistAndDontOverwrite(exportPath):
            return None

        self.logger.info("Printing the TRNSYS file...")

        if exportTo == "mfs":
            header = open(os.path.join(ddckFolder, "generic", "head.ddck"), "r", encoding="windows-1252")
            headerLines = header.readlines()
            for line in headerLines:
                if line[:4] == "STOP":
                    fullExportText += "STOP = 1 \n"
                else:
                    fullExportText += line
            header.close()
        elif exportTo == "ddck":
            fullExportText += "*************************************\n"
            fullExportText += "** BEGIN hydraulic.ddck\n"
            fullExportText += "*************************************\n\n"
            fullExportText += "*************************************\n"
            fullExportText += "** Outputs to energy balance in kWh\n"
            fullExportText += (
                "** Following this naming standard : qSysIn_name, qSysOut_name, elSysIn_name, elSysOut_name\n"
            )
            fullExportText += "*************************************\n"
            fullExportText += "EQUATIONS 1\n"
            fullExportText += "qSysOut_PipeLoss = PipeLossTot\n"

        simulationUnit = 450
        simulationType = 935
        descConnLength = 15

        exporter = self._createExporter()

        blackBoxProblem, blackBoxText = exporter.exportBlackBox(exportTo=exportTo)
        if blackBoxProblem:
            return None

        fullExportText += blackBoxText
        if exportTo == "mfs":
            fullExportText += exporter.exportMassFlows()
            fullExportText += exporter.exportPumpOutlets()
            fullExportText += exporter.exportDivSetting(simulationUnit - 10)

        fullExportText += exporter.exportDoublePipeParameters(exportTo=exportTo)

        fullExportText += exporter.exportParametersFlowSolver(simulationUnit, simulationType, descConnLength)

        fullExportText += exporter.exportInputsFlowSolver()
        fullExportText += exporter.exportOutputsFlowSolver(simulationUnit)
        fullExportText += exporter.exportFluids() + "\n"
        fullExportText += exporter.exportHydraulicLoops() + "\n"
        fullExportText += exporter.exportPipeAndTeeTypesForTemp(simulationUnit + 1)  # DC-ERROR
        fullExportText += exporter.exportPrintPipeLosses()

        fullExportText += exporter.exportMassFlowPrinter(self.printerUnitnr, 15)
        fullExportText += exporter.exportTempPrinter(self.printerUnitnr + 1, 15)

        if exportTo == "mfs":
            fullExportText += "CONSTANTS 1\nTRoomStore=1\n"
            fullExportText += "ENDS"

        self.logger.info("------------------------> END OF EXPORT <------------------------")

        if exportTo == "mfs":
            f = open(exportPath, "w")
            f.truncate(0)
            f.write(fullExportText)
            f.close()
        elif exportTo == "ddck":
            if fullExportText[:1] == "\n":
                fullExportText = fullExportText[1:]
            hydraulicFolder = os.path.split(exportPath)[0]
            if not (os.path.isdir(hydraulicFolder)):
                os.makedirs(hydraulicFolder)
            f = open(exportPath, "w")
            f.truncate(0)
            f.write(fullExportText)
            f.close()

        try:
            lines = _du.loadDeck(exportPath, eraseBeginComment=True, eliminateComments=True)
            _du.checkEquationsAndConstants(lines, exportPath)
        except Exception as error:
            errorMessage = f"An error occurred while exporting the system hydraulics: {error}"
            _errs.showErrorMessageBox(errorMessage)
            return None

        return exportPath

    def _createExporter(self) -> Export:
        massFlowContributors = self._getMassFlowContributors()
        exporter = Export(
            self.diagramName,
            massFlowContributors,
            self.hydraulicLoops.hydraulicLoops,
            self.fluids.fluids,
            self.logger,
            self,
        )
        return exporter

    def _getMassFlowContributors(self) -> _tp.Sequence[_mfs.MassFlowNetworkContributorMixin]:
        massFlowContributors = [o for o in self.trnsysObj if isinstance(o, _mfs.MassFlowNetworkContributorMixin)]
        return massFlowContributors

    def _isHydraulicConnected(self) -> bool:
        for obj in self.trnsysObj:
            if not isinstance(obj, _mfs.MassFlowNetworkContributorMixin):
                continue

            internalPiping = obj.getInternalPiping()

            for portItem in internalPiping.modelPortItemsToGraphicalPortItem.values():
                if not portItem.connectionList:
                    return False

        return True

    def _doesFileExistAndDontOverwrite(self, folderPath):
        if not _pl.Path(folderPath).exists():
            return False

        qmb = QMessageBox(self)
        qmb.setText(f"Warning: {folderPath} already exists. Do you want to overwrite it or cancel?")
        qmb.setStandardButtons(QMessageBox.Save | QMessageBox.Cancel)
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()

        if ret == QMessageBox.Cancel:
            self.canceled = True
            self.logger.info("Canceling")
            return True

        self.canceled = False
        self.logger.info("Overwriting")
        return False

    def exportHydraulicControl(self):
        self.logger.info("------------------------> START OF EXPORT <------------------------")

        self.sortTrnsysObj()

        fullExportText = ""

        ddckFolder = os.path.join(self.projectFolder, "ddck")

        hydCtrlPath = os.path.join(ddckFolder, "control", "hydraulic_control.ddck")
        if _pl.Path(hydCtrlPath).exists():
            qmb = QMessageBox(self)
            qmb.setText(
                "Warning: "
                + "The file hydraulic_control.ddck already exists in the control folder. Do you want to overwrite it or cancel?"
            )
            qmb.setStandardButtons(QMessageBox.Save | QMessageBox.Cancel)
            qmb.setDefaultButton(QMessageBox.Cancel)
            ret = qmb.exec()
            if ret == QMessageBox.Save:
                self.canceled = False
                self.logger.info("Overwriting")
            else:
                self.canceled = True
                self.logger.info("Canceling")
                return

        fullExportText += "*************************************\n"
        fullExportText += "**BEGIN hydraulic_control.ddck\n"
        fullExportText += "*************************************\n"

        simulationUnit = 450

        exporter = self._createExporter()

        fullExportText += exporter.exportPumpOutlets()
        fullExportText += exporter.exportMassFlows()
        fullExportText += exporter.exportDivSetting(simulationUnit - 10)

        self.logger.info("------------------------> END OF EXPORT <------------------------")

        if fullExportText[:1] == "\n":
            fullExportText = fullExportText[1:]
        controlFolder = os.path.split(hydCtrlPath)[0]
        if not (os.path.isdir(controlFolder)):
            os.makedirs(controlFolder)
        f = open(str(hydCtrlPath), "w")
        f.truncate(0)
        f.write(fullExportText)
        f.close()

        return hydCtrlPath

    def sortTrnsysObj(self):
        res = self.trnsysObj.sort(key=self.sortId)
        for s in self.trnsysObj:
            self.logger.debug("s has tr id " + str(s.trnsysId) + " has dname " + s.displayName)

    def sortId(self, l1):
        """
        Sort function returning a sortable key
        Parameters
        ----------
        l1 : Block/Connection

        Returns
        -------

        """
        return l1.trnsysId

    def setName(self, newName):
        self.diagramName = newName

    def delBlocks(self):
        """
        Deletes the whole diagram

        Returns
        -------

        """
        self.hydraulicLoops.clear()

        while len(self.trnsysObj) > 0:
            self.logger.info("In deleting...")
            self.trnsysObj[0].deleteBlock()

        while len(self.graphicalObj) > 0:
            self.graphicalObj[0].deleteBlock()

    # Encoding / decoding
    def encodeDiagram(self, filename):
        """
        Encodes the diagram to a json file.

        Parameters
        ----------
        filename : str

        Returns
        -------

        """
        self.logger.info("filename is at encoder " + str(filename))

        with open(filename, "w") as jsonfile:
            json.dump(self, jsonfile, indent=4, sort_keys=True, cls=_enc.Encoder)

    def _decodeDiagram(self, filename, loadValue="load"):
        self.logger.info("Decoding " + filename)
        with open(filename, "r") as jsonfile:
            blocklist = json.load(jsonfile, cls=Decoder, editor=self)

        blockFolderNames = []

        for j in blocklist["Blocks"]:
            for k in j:
                if isinstance(k, BlockItem):
                    k.setParent(self.diagramView)
                    k.changeSize()
                    self.diagramScene.addItem(k)
                    blockFolderNames.append(k.displayName)

                if isinstance(k, StorageTank):
                    k.updateImage()

                if isinstance(k, GraphicalItem):
                    k.setParent(self.diagramView)
                    self.diagramScene.addItem(k)

                if isinstance(k, dict):
                    if "__idDct__" in k:
                        # here we don't set the ids because the copyGroup would need access to idGen
                        self.logger.debug("Found the id dict while loading, not setting the ids")

                        self.idGen.setID(k["GlobalId"])
                        self.idGen.setTrnsysID(k["trnsysID"])
                        self.idGen.setConnID(k["globalConnID"])

                    if "__nameDct__" in k:
                        self.logger.debug("Found the name dict while loading")
                        if loadValue == "load":
                            self.diagramName = k["DiagramName"]

        blockFolderNames.append("generic")
        blockFolderNames.append("hydraulic")
        blockFolderNames.append("weather")
        blockFolderNames.append("control")

        ddckFolder = os.path.join(self.projectFolder, "ddck")
        ddckFolders = os.listdir(ddckFolder)
        additionalFolders = []

        for folder in ddckFolders:
            if folder not in blockFolderNames and "StorageTank" not in folder:
                additionalFolders.append(folder)

        if len(additionalFolders) > 0:
            warnBox = QMessageBox()
            warnBox.setWindowTitle("Additional ddck-folders")

            if len(additionalFolders) == 1:
                text = "The following ddck-folder does not have a corresponding component in the diagram:"
            else:
                text = "The following ddck-folders do not have a corresponding component in the diagram:"

            for folder in additionalFolders:
                text += "\n\t" + folder

            warnBox.setText(text)
            warnBox.setStandardButtons(QMessageBox.Ok)
            warnBox.setDefaultButton(QMessageBox.Ok)
            warnBox.exec()

        for t in self.trnsysObj:
            t.assignIDsToUninitializedValuesAfterJsonFormatMigration(self.idGen)

            self.logger.debug("Tr obj is" + str(t) + " " + str(t.trnsysId))
            if hasattr(t, "isTempering"):
                self.logger.debug("tv has " + str(t.isTempering))

        self._decodeHydraulicLoops(blocklist)

    def _decodeHydraulicLoops(self, blocklist):
        singlePipeConnections = [c for c in self.connectionList if isinstance(c, SinglePipeConnection)]
        if "hydraulicLoops" not in blocklist:
            hydraulicLoops = _hlmig.createLoops(singlePipeConnections, self.fluids.WATER)
        else:
            serializedHydraulicLoops = blocklist["hydraulicLoops"]
            hydraulicLoops = _hlm.HydraulicLoops.createFromJson(
                serializedHydraulicLoops, singlePipeConnections, self.fluids
            )

        self.hydraulicLoops = hydraulicLoops

    def exportSvg(self):
        """
        For exporting a svg file (text is still too large)
        Returns
        -------

        """
        generator = QSvgGenerator()
        generator.setResolution(300)
        generator.setSize(QSize(self.diagramScene.width(), self.diagramScene.height()))
        # generator.setViewBox(QRect(0, 0, 800, 800))
        generator.setViewBox(self.diagramScene.sceneRect())
        generator.setFileName("VectorGraphicsExport.svg")

        painter = QPainter()
        painter.begin(generator)
        painter.setRenderHint(QPainter.Antialiasing)
        self.diagramScene.render(painter)
        painter.end()

    def exportDdckPlaceHolderValuesJsonFile(self) -> _res.Result[None]:
        if not self._isHydraulicConnected():
            return _res.Error(f"You need to connect all port items before you can export the hydraulics.")

        jsonFilePath = _pl.Path(self.projectFolder) / "DdckPlaceHolderValues.json"

        if jsonFilePath.is_dir():
            qmb = QMessageBox(self)
            qmb.setText(
                f"A folder already exits at f{jsonFilePath}. Chose a different location or delete the folder first."
            )
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.exec()

            return None

        if jsonFilePath.is_file():
            qmb = QMessageBox(self)
            qmb.setText("The file already exists. Do you want to overwrite it or cancel?")
            qmb.setStandardButtons(QMessageBox.Save | QMessageBox.Cancel)
            qmb.setDefaultButton(QMessageBox.Cancel)
            ret = qmb.exec()

            if ret != QMessageBox.Save:
                return None

        result = self.encodeDdckPlaceHolderValuesToJson(jsonFilePath)
        if _res.isError(result):
            return _res.error(result)

        msgb = QMessageBox(self)
        msgb.setText(f"Saved place holder values JSON file at {jsonFilePath}.")
        msgb.exec()

        return None

    def encodeDdckPlaceHolderValuesToJson(self, filePath: _pl.Path) -> _res.Result[None]:
        ddckDirNames = self._getDdckDirNames()

        result = _ph.getPlaceholderValues(ddckDirNames, self.trnsysObj)
        if _res.isError(result):
            return _res.error(result)

        ddckPlaceHolderValuesDictionary = _res.value(result)

        jsonContent = json.dumps(ddckPlaceHolderValuesDictionary, indent=4, sort_keys=True)
        filePath.write_text(jsonContent)

    # Saving related
    def save(self, showWarning=True):
        """
        If saveas has not been used, diagram will be saved in "/diagrams"
        If saveas has been used, diagram will be saved in self.saveAsPath
        Returns
        -------

        """
        self.diagramName = os.path.split(self.projectFolder)[-1] + ".json"
        diagramPath = os.path.join(self.projectFolder, self.diagramName)

        if os.path.isfile(diagramPath) and showWarning:
            qmb = QMessageBox(self)
            qmb.setText("Warning: " + "This diagram name exists already. Do you want to overwrite or cancel?")
            qmb.setStandardButtons(QMessageBox.Save | QMessageBox.Cancel)
            qmb.setDefaultButton(QMessageBox.Cancel)
            ret = qmb.exec()

            if ret != QMessageBox.Save:
                self.logger.info("Canceling")
                return

            self.logger.info("Overwriting")
            self.encodeDiagram(diagramPath)

        self.encodeDiagram(diagramPath)
        msgb = QMessageBox(self)
        msgb.setText("Saved diagram at " + diagramPath)
        msgb.exec()

    def saveToProject(self):
        projectPath = self.projectPath

    def renameDiagram(self, newName):
        """

        Parameters
        ----------
        newName

        Returns
        -------

        """

        if self.saveAsPath.name != "":
            # print("Path name is " + self.saveAsPath.name)
            if newName + ".json" in self.saveAsPath.glob("*"):
                QMessageBox(
                    self, "Warning", "This diagram name exists already in the directory." " Please rename this diagram"
                )
            else:
                self.saveAsPath = _pl.Path(
                    self.saveAsPath.stem[0 : self.saveAsPath.name.index(self.diagramName)] + newName
                )

        self.diagramName = newName
        self.parent().currentFile = newName
        # fromPath = self.projectFolder
        # destPath = os.path.dirname(__file__)
        # destPath = os.path.join(destPath, 'default')
        # destPath = os.path.join(destPath, newName)
        # os.rename(fromPath, destPath)

        # print("Path is now: " + str(self.saveAsPath))
        # print("Diagram name is: " + self.diagramName)

    def saveAtClose(self):
        self.logger.info("saveaspath is " + str(self.saveAsPath))

        # closeDialog = closeDlg()
        # if closeDialog.closeBool:
        filepath = _pl.Path(_pl.Path(__file__).resolve().parent.joinpath("recent"))
        self.encodeDiagram(str(filepath.joinpath(self.diagramName + ".json")))

    # Mode related
    def setAlignMode(self, b):
        self.alignMode = True

    def setEditorMode(self, b):
        self.editorMode = b

    def setMoveDirectPorts(self, b):
        """
        Sets the bool moveDirectPorts. When mouse released in diagramScene, moveDirectPorts is set to False again
        Parameters
        ----------
        b : bool

        Returns
        -------

        """
        self.moveDirectPorts = b

    def setSnapGrid(self, b):
        self.snapGrid = b

    def setSnapSize(self, s):
        self.snapSize = s

    def setConnLabelVis(self, isVisible: bool) -> None:
        for c in self.trnsysObj:
            if isinstance(c, ConnectionBase):
                c.setLabelVisible(isVisible)
            if isinstance(c, BlockItem):
                c.label.setVisible(isVisible)
            if isinstance(c, TVentil):
                c.posLabel.setVisible(isVisible)

    def updateConnGrads(self):
        for t in self.trnsysObj:
            if isinstance(t, ConnectionBase):
                t.updateSegGrads()

    # Dialog calls
    def showBlockDlg(self, bl):
        c = BlockDlg(bl, self)

    def showDoublePipeBlockDlg(self, bl):
        c = DoublePipeBlockDlg(bl, self)

    def showPumpDlg(self, bl):
        c = PumpDlg(bl, self)

    def showDiagramDlg(self):
        c = diagramDlg(self)

    def showGenericPortPairDlg(self, bl):
        c = GenericPortPairDlg(bl, self)

    def showHxDlg(self, hx):
        c = hxDlg(hx, self)

    def showSegmentDlg(self, seg):
        c = segmentDlg(seg, self)

    def showTVentilDlg(self, bl):
        c = TVentilDlg(bl, self)

    def showConfigStorageDlg(self, bl):
        c = ConfigureStorageDialog(bl, self)

    def getConnection(self, n):
        return self.connectionList[int(n)]

    # Unused
    def create_icon(self, map_icon):
        map_icon.fill()
        painter = QPainter(map_icon)
        painter.fillRect(10, 10, 40, 40, QColor(88, 233, 252))
        # painter.setBrush(Qt.red)
        painter.setBrush(QColor(252, 136, 98))
        painter.drawEllipse(36, 2, 15, 15)
        painter.setBrush(Qt.yellow)
        painter.drawEllipse(20, 20, 20, 20)
        painter.end()

    def setTrnsysIdBack(self):
        self.idGen.trnsysID = max(t.trnsysId for t in self.trnsysObj)

    def findStorageCorrespPorts1(self, portList):
        """
        This function gets the ports on the other side of pipes connected to a port of the StorageTank. Unused

        Parameters
        ----------
        portList : :obj:`List` of :obj:`PortItems`

        Returns
        -------

        """

        res = []
        # print("Finding c ports")
        for p in portList:
            if len(p.connectionList) > 0:  # check if not >1 needed
                # connectionList[0] is the hidden connection created when the portPair is
                i = 0
                # while type(p.connectionList[i].fromPort.parent) is StorageTank and type(p.connectionList[i].toPort.parent) is StorageTank:
                while (p.connectionList[i].fromPort.parent) == (p.connectionList[i].toPort.parent):
                    i += 1
                if len(p.connectionList) >= i + 1:
                    if p.connectionList[i].fromPort is p:
                        res.append(p.connectionList[i].toPort)
                    elif p.connectionList[i].toPort is p:
                        res.append(p.connectionList[i].fromPort)
                    else:
                        self.logger.debug("Port is not fromPort nor toPort")

        # [print(p.parent.displayName) for p in res]
        return res

    def printPDF(self):
        """
        ---------------------------------------------
        Export diagram as pdf onto specified folder
        fn = user input directory
        ---------------------------------------------
        """
        fn, _ = QFileDialog.getSaveFileName(self, "Export PDF", None, "PDF files (.pdf);;All Files()")
        if fn != "":
            if QFileInfo(fn).suffix() == "":
                fn += ".pdf"
            printer = QPrinter(QPrinter.HighResolution)
            printer.setOrientation(QPrinter.Landscape)
            printer.setOutputFormat(QPrinter.PdfFormat)
            printer.setOutputFileName(fn)
            painter = QPainter(printer)
            self.diagramScene.render(painter)
            painter.end()
            self.logger.info("File exported to %s" % fn)

    def openProject(self):
        self.projectPath = str(QFileDialog.getExistingDirectory(self, "Select Project Path"))
        if self.projectPath != "":
            test = self.parent()

            self.parent().newDia()
            self.PPL.setText(self.projectPath)
            loadPath = os.path.join(self.projectPath, "ddck")

            self.createConfigBrowser(self.projectPath)
            self.copyGenericFolder(self.projectPath)
            self.createHydraulicDir(self.projectPath)
            self.createWeatherAndControlDirs(self.projectPath)
            self.createDdckTree(loadPath)
            # todo : open diagram
            # todo : add files into list

    def createDdckTree(self, loadPath):
        treeToRemove = self.findChild(QTreeView, "ddck")
        try:
            # treeToRemove.hide()
            treeToRemove.deleteLater()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")
        if self.projectPath == "":
            loadPath = os.path.join(loadPath, "ddck")
        if not os.path.exists(loadPath):
            os.makedirs(loadPath)
        self.model = MyQFileSystemModel()
        self.model.setRootPath(loadPath)
        self.model.setName("ddck")
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(loadPath))
        self.tree.setObjectName("ddck")
        self.tree.setMinimumHeight(600)
        self.tree.setSortingEnabled(True)
        self.splitter.insertWidget(0, self.tree)

    def createConfigBrowser(self, loadPath):
        self.layoutToRemove = self.findChild(QHBoxLayout, "Config_Layout")
        try:
            # treeToRemove.hide()
            self.layoutToRemove.deleteLater()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")

        runConfigData = self._getPackageResourceData("templates/run.config")
        runConfigPath = _pl.Path(loadPath) / "run.config"
        runConfigPath.write_bytes(runConfigData)

        self.HBox = QHBoxLayout()
        self.refreshButton = QPushButton(self)
        self.refreshButton.setIcon(_img.ROTATE_TO_RIGHT_PNG.icon())
        self.refreshButton.clicked.connect(self.refreshConfig)
        self.model = MyQFileSystemModel()
        self.model.setRootPath(loadPath)
        self.model.setName("Config File")
        self.model.setFilter(QDir.Files)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(loadPath))
        self.tree.setObjectName("config")
        self.tree.setFixedHeight(60)
        self.tree.setSortingEnabled(False)
        self.HBox.addWidget(self.refreshButton)
        self.HBox.addWidget(self.tree)
        self.HBox.setObjectName("Config_Layout")
        self.fileBrowserLayout.addLayout(self.HBox)

    def createProjectFolder(self):
        if not os.path.exists(self.projectFolder):
            os.makedirs(self.projectFolder)

    def refreshConfig(self):
        # configPath = os.path.dirname(__file__)
        # configPath = os.path.join(configPath, 'project')
        # configPath = os.path.join(configPath, self.date_time)
        # emptyConfig = os.path.join(configPath, 'run.config')
        if self.projectPath == "":
            localPath = self.projectFolder
        else:
            localPath = self.projectPath

        self.configToEdit = os.path.join(localPath, "run.config")
        os.remove(self.configToEdit)
        shutil.copy(self.emptyConfig, localPath)
        self.configToEdit = os.path.join(localPath, "run.config")

        localDdckPath = os.path.join(localPath, "ddck")
        with open(self.configToEdit, "r") as file:
            lines = file.readlines()
        localPathStr = "string LOCAL$ %s" % str(localDdckPath)
        lines[21] = localPathStr + "\n"

        with open(self.configToEdit, "w") as file:
            file.writelines(lines)

        # print(localPathStr)
        self.userInputList()

    def userInputList(self):
        self.logger.debug(self.fileList)
        dia = FileOrderingDialog(self.fileList, self)

    def copyGenericFolder(self, loadPath):
        genericFolderPath = _pl.Path(loadPath) / "ddck" / "generic"

        if not genericFolderPath.exists():
            self.logger.info("Creating %s", genericFolderPath)
            genericFolderPath.mkdir()

        headData = self._getPackageResourceData("templates/generic/head.ddck")
        self.logger.info("Copying head.ddck")
        (genericFolderPath / "head.ddck").write_bytes(headData)

        endData = self._getPackageResourceData("templates/generic/end.ddck")
        self.logger.info("Copying end.ddck")
        (genericFolderPath / "end.ddck").write_bytes(endData)

    @staticmethod
    def _getPackageResourceData(resourcePath):
        data = _pu.get_data(_tgui.__name__, resourcePath)
        assert data, f"{resourcePath} package resource not found"
        return data

    def createHydraulicDir(self, projectPath):

        self.hydraulicFolder = os.path.join(projectPath, "ddck")
        self.hydraulicFolder = os.path.join(self.hydraulicFolder, "hydraulic")

        if not os.path.exists(self.hydraulicFolder):
            self.logger.info("Creating " + self.hydraulicFolder)
            os.makedirs(self.hydraulicFolder)

    def createWeatherAndControlDirs(self, projectPath):

        ddckFolder = os.path.join(projectPath, "ddck")
        weatherFolder = os.path.join(ddckFolder, "weather")
        controlFolder = os.path.join(ddckFolder, "control")

        if not os.path.exists(weatherFolder):
            self.logger.info("Creating " + weatherFolder)
            os.makedirs(weatherFolder)

        if not os.path.exists(controlFolder):
            self.logger.info("Creating " + controlFolder)
            os.makedirs(controlFolder)

    def editHydraulicLoop(self, singlePipeConnection: SinglePipeConnection):
        assert isinstance(singlePipeConnection.fromPort, SinglePipePortItem)

        hydraulicLoop = self.hydraulicLoops.getLoopForExistingConnection(singlePipeConnection)
        _hledit.edit(hydraulicLoop, self.hydraulicLoops, self.fluids)

    def _getDdckDirNames(self) -> _tp.Sequence[str]:
        ddckDirPath = _pl.Path(self.projectFolder) / "ddck"

        componentDdckDirPaths = list(ddckDirPath.iterdir())

        ddckDirNames = []
        for componentDirPath in componentDdckDirPaths:
            ddckDirNames.append(componentDirPath.name)

        return ddckDirNames

    def nameExists(self, name):
        for item in self.trnsysObj:
            if str(item.displayName).lower() == name.lower():
                return True
        return False

    def nameExistsInDdckFolder(self, name):
        projectFolderDdckPath = _pl.Path(self.projectFolder) / "ddck"
        projectDdckFiles = projectFolderDdckPath.iterdir()
        for file in projectDdckFiles:
            if file.name.lower() == name.lower():
                return True
        return False



# pylint: disable = invalid-name

import glob
import os

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class BlockItemFourPorts(BlockItem, MassFlowNetworkContributorMixin):  # pylint: disable = too-many-instance-attributes
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.logger = parent.logger

        self.h = 120
        self.w = 120
        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.inputs.append(_cspi.createSinglePipePortItem("i", 2, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 0, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))

        self.childIds = []
        self.childIds.append(self.trnsysId)
        self.childIds.append(self.parent.parent().idGen.getTrnsysID())

        self.changeSize()

    def encode(self):
        if not self.isVisible():
            return None
        self.logger.debug(f"Encoding a {self.name} block")

        portListInputs = []
        portListOutputs = []

        for inputPort in self.inputs:
            portListInputs.append(inputPort.id)
        for outputPort in self.outputs:
            portListOutputs.append(outputPort.id)

        dct = {}
        dct[".__BlockDict__"] = True
        dct["BlockName"] = self.name
        dct["BlockDisplayName"] = self.displayName
        dct["PortsIDIn"] = portListInputs
        dct["PortsIDOut"] = portListOutputs
        dct[self.name + "Position"] = (float(self.pos().x()), float(self.pos().y()))
        dct["ID"] = self.id
        dct["trnsysID"] = self.trnsysId
        dct["childIds"] = self.childIds
        dct["FlippedH"] = self.flippedH
        dct["FlippedV"] = self.flippedV
        dct["RotationN"] = self.rotationN

        dictName = "Block-"

        return dictName, dct

    def decode(self, i, resBlockList):

        self.flippedH = i["FlippedH"]
        self.flippedV = i["FlippedV"]
        self.childIds = i["childIds"]
        self.displayName = i["BlockDisplayName"]
        self.changeSize()

        for x, inputPort in enumerate(self.inputs):
            inputPort.id = i["PortsIDIn"][x]

        for x, outputPort in enumerate(self.outputs):
            outputPort.id = i["PortsIDOut"][x]

        self.setPos(float(i[self.name + "Position"][0]), float(i[self.name + "Position"][1]))
        self.trnsysId = i["trnsysID"]
        self.id = i["ID"]

        resBlockList.append(self)

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs):

        self.changeSize()

        for x, inputPort in enumerate(self.inputs):
            inputPort.id = i["PortsIDIn"][x]

        for x, outputPort in enumerate(self.outputs):
            outputPort.id = i["PortsIDOut"][x]

        self.setPos(float(i[self.name + "Position"][0]) + offset_x, float(i[self.name + "Position"][1] + offset_y))

        resBlockList.append(self)

    def exportBlackBox(self):
        equations = []
        files = glob.glob(os.path.join(self.path, "**/*.ddck"), recursive=True)
        if not files:
            status = "noDdckFile"
        else:
            status = "noDdckEntry"
        lines = []
        for file in files:
            with open(file, "r") as infile:  # pylint: disable = unspecified-encoding
                lines += infile.readlines()
        for i, line in enumerate(lines):
            if "output" in line.lower() and "to" in line.lower() and "hydraulic" in line.lower():
                counter = 1
                for j in range(i, len(lines) - i):
                    if lines[j][0] == "T":
                        outputT = lines[j].split("=")[0].replace(" ", "")
                        equations.append("T" + self.displayName + "X" + str(counter) + "=1 ! suggestion: " + outputT)
                        counter += 1
                    if counter == 3:
                        status = "success"
                        break
                break

        if status in ("noDdckFile", "noDdckEntry"):
            equations.append("T" + self.displayName + "X1" + "=1")
            equations.append("T" + self.displayName + "X2" + "=1")

        return status, equations

    def getInternalPiping(self) -> InternalPiping:
        side1Input = _mfn.PortItem("Side Input 1", _mfn.PortItemType.INPUT)
        side1Output = _mfn.PortItem("Side Output 1", _mfn.PortItemType.OUTPUT)
        side1Pipe = _mfn.Pipe(f"{self.displayName}Side1", self.childIds[0], side1Input, side1Output)

        side2Input = _mfn.PortItem("Side Input 2", _mfn.PortItemType.INPUT)
        side2Output = _mfn.PortItem("Side Output 2", _mfn.PortItemType.OUTPUT)
        side2Pipe = _mfn.Pipe(f"{self.displayName}Side2", self.childIds[1], side2Input, side2Output)

        modelPortItemsToGraphicalPortItem = {
            side1Input: self.inputs[0],
            side1Output: self.outputs[0],
            side2Input: self.inputs[1],
            side2Output: self.outputs[1]
        }

        return InternalPiping([side1Pipe, side2Pipe], modelPortItemsToGraphicalPortItem)

    def getSubBlockOffset(self, c):  # pylint: disable = invalid-name
        for i in range(2):
            if (
                    self.inputs[i] == c.toPort
                    or self.inputs[i] == c.fromPort
                    or self.outputs[i] == c.toPort
                    or self.outputs[i] == c.fromPort
            ):
                return i
        return None



from __future__ import annotations

import typing as tp

from PyQt5 import QtGui, QtCore
from PyQt5.QtCore import QPointF
from PyQt5.QtGui import QLinearGradient, QBrush, QPen
from PyQt5.QtWidgets import QGraphicsLineItem, QMenu

from trnsysGUI.CornerItem import CornerItem  # type: ignore[attr-defined]
from trnsysGUI.SegmentItemBase import SegmentItemBase  # type: ignore[attr-defined]

# This is needed to avoid a circular import but still be able to type check
if tp.TYPE_CHECKING:
    from trnsysGUI.connection.singlePipeConnection import SinglePipeConnection  # type: ignore[attr-defined]  # pylint: disable=unused-import


class SinglePipeSegmentItem(SegmentItemBase):
    def __init__(self, startNode, endNode, parent: SinglePipeConnection):
        super().__init__(startNode, endNode, parent)

        self._singlePipeConnection = parent

        self.singleLine = QGraphicsLineItem(self)
        self.linearGrad = None
        self.initGrad()

    def _createSegment(self, startNode, endNode) -> "SegmentItemBase":
        return SinglePipeSegmentItem(startNode, endNode, self._singlePipeConnection)

    def _getContextMenu(self) -> QMenu:
        menu = super()._getContextMenu()

        editHydraulicLoopAction = menu.addAction("Edit hydraulic loop")

        editHydraulicLoopAction.triggered.connect(self._singlePipeConnection.editHydraulicLoop)

        return menu

    def _setLineImpl(self, x1, y1, x2, y2):
        self.initGrad()
        self.singleLine.setLine(x1, y1, x2, y2)
        self.linePoints = self.singleLine.line()

    def initGrad(self):
        color = QtCore.Qt.white

        pen1 = QtGui.QPen(color, 4)

        if isinstance(self.startNode.parent, CornerItem):
            startBlock = self.startNode.firstNode().parent
        else:
            startBlock = self.startNode.parent

        if isinstance(self.endNode.parent, CornerItem):
            endBlock = self.endNode.lastNode().parent
        else:
            endBlock = self.endNode.parent

        self.linearGrad = QLinearGradient(
            QPointF(startBlock.fromPort.scenePos().x(), startBlock.fromPort.scenePos().y()),
            QPointF(endBlock.toPort.scenePos().x(), endBlock.toPort.scenePos().y()),
        )
        self.linearGrad.setColorAt(0, QtCore.Qt.blue)
        self.linearGrad.setColorAt(1, QtCore.Qt.red)

        self.linearGrad.setColorAt(0, QtCore.Qt.gray)
        self.linearGrad.setColorAt(1, QtCore.Qt.black)

        pen1.setBrush(QBrush(self.linearGrad))

        self.singleLine.setPen(pen1)

    def updateGrad(self):
        color = QtCore.Qt.white
        pen1 = QtGui.QPen(color, 4)

        totLenConn = self.connection.totalLength()
        partLen1 = self.connection.partialLength(self.startNode)
        partLen2 = self.connection.partialLength(self.endNode)

        if isinstance(self.startNode.parent, CornerItem):
            startGradP = QPointF(self.startNode.parent.scenePos().x(), self.startNode.parent.scenePos().y())
        elif self.startNode.prevN() is None:
            startGradP = QPointF(
                self.startNode.parent.fromPort.scenePos().x(), self.startNode.parent.fromPort.scenePos().y()
            )
        else:
            startGradP = QPointF(self.line().p1().x(), self.line().p1().y())

        if isinstance(self.endNode.parent, CornerItem):
            endGradP = QPointF(self.endNode.parent.scenePos().x(), self.endNode.parent.scenePos().y())
        elif self.endNode.nextN() is None:
            endGradP = QPointF(self.endNode.parent.toPort.scenePos().x(), self.endNode.parent.toPort.scenePos().y())
        else:
            endGradP = QPointF(self.line().p2().x(), self.line().p2().y())

        self.linearGrad = QLinearGradient(startGradP, endGradP)

        self.linearGrad.setColorAt(0, self.interpolate(partLen1, totLenConn))
        self.linearGrad.setColorAt(1, self.interpolate(partLen2, totLenConn))

        pen1.setBrush(QBrush(self.linearGrad))

        self.singleLine.setPen(pen1)

    def setSelect(self, isSelected: bool) -> None:
        if isSelected:
            selectPen = self._createSelectPen()
            self.singleLine.setPen(selectPen)
        else:
            self.updateGrad()

    def setColorAndWidthAccordingToMassflow(self, color, width):
        pen1 = QPen(color, width)
        self.singleLine.setPen(pen1)



# pylint: disable=invalid-name

import glob
import os
import typing as _tp

from PyQt5 import QtCore
from PyQt5.QtCore import QPointF, QEvent, QTimer
from PyQt5.QtGui import QPixmap, QCursor, QMouseEvent
from PyQt5.QtWidgets import QGraphicsPixmapItem, QGraphicsTextItem, QMenu, QTreeView

import trnsysGUI.images as _img
from trnsysGUI import idGenerator as _id
from trnsysGUI.MoveCommand import MoveCommand  # type: ignore[attr-defined]
from trnsysGUI.ResizerItem import ResizerItem  # type: ignore[attr-defined]
from trnsysGUI.blockItemModel import BlockItemModel
from trnsysGUI.doublePipePortItem import DoublePipePortItem
from trnsysGUI.singlePipePortItem import SinglePipePortItem

FILEPATH = "res/Config.txt"


# pylint: disable = fixme
# TODO : TeePiece and AirSourceHp size ratio need to be fixed, maybe just use original
#  svg instead of modified ones, TVentil is flipped. heatExchangers are also wrongly oriented
class BlockItem(QGraphicsPixmapItem):  # pylint: disable = too-many-public-methods, too-many-instance-attributes
    def __init__(self, trnsysType, parent, displayNamePrefix=None, displayName=None, **kwargs):
        super().__init__(None)

        self.logger = parent.logger

        self.w = 120
        self.h = 120
        self.parent = parent
        self.id = self.parent.parent().idGen.getID()
        self.propertyFile = []

        if displayNamePrefix:
            self.displayName = displayNamePrefix + str(self.id)
        elif displayName:
            self.displayName = displayName
        else:
            raise Exception('No display name defined.')

        if "loadedBlock" not in kwargs:
            self.parent.parent().trnsysObj.append(self)

        self.inputs = []
        self.outputs = []

        self.path = None

        # Export related:
        self.name = trnsysType
        self.trnsysId = self.parent.parent().idGen.getTrnsysID()

        # Transform related
        self.flippedV = False
        self.flippedH = False
        self.rotationN = 0
        self.flippedHInt = -1
        self.flippedVInt = -1

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

        # To set flags of this item
        self.setFlags(self.ItemIsSelectable | self.ItemIsMovable)
        self.setFlag(self.ItemSendsScenePositionChanges, True)
        self.setCursor(QCursor(QtCore.Qt.PointingHandCursor))

        self.label = QGraphicsTextItem(self.displayName, self)
        self.label.setVisible(False)

        # Experimental, used for detecting genereated blocks attached to storage ports
        self.inFirstRow = False

        # Undo framework related
        self.oldPos = None

        self.origOutputsPos = None
        self.origInputsPos = None

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        if isinstance(self, BlockItem):
            raise AssertionError("`BlockItem' cannot be instantiated directly.")

        currentClassName = BlockItem.__name__
        currentMethodName = f"{currentClassName}.{BlockItem._getImageAccessor.__name__}"

        message = (
            f"{currentMethodName} has been called. However, this method should not be called directly but must\n"
            f"implemented in a child class. This means that a) someone instantiated `{currentClassName}` directly\n"
            f"or b) a child class of it doesn't implement `{currentMethodName}`. Either way that's an\n"
            f"unrecoverable error and therefore the program will be terminated now. Please do get in touch with\n"
            f"the developers if you've encountered this error. Thanks."
        )

        exception = AssertionError(message)

        # I've seen exception messages mysteriously swallowed that's why we're logging the message here, too.
        self.logger.error(message, exc_info=exception, stack_info=True)

        raise exception

    def addTree(self):
        pass

    def hasDdckPlaceHolders(self):  # pylint: disable = no-self-use
        return False

    # Setter functions
    def setParent(self, p):
        self.parent = p

        if self not in self.parent.parent().trnsysObj:
            self.parent.parent().trnsysObj.append(self)
            # self.logger.debug("trnsysObj are " + str(self.parent.parent().trnsysObj))

    def setId(self, newId):
        self.id = newId

    def setDisplayName(self, newName: str) -> None:
        self.displayName = newName
        self.label.setPlainText(newName)

    # Interaction related
    def contextMenuEvent(self, event):
        menu = QMenu()

        a1 = menu.addAction("Launch NotePad++")
        a1.triggered.connect(self.launchNotepadFile)

        rr = _img.ROTATE_TO_RIGHT_PNG.icon()
        a2 = menu.addAction(rr, "Rotate Block clockwise")
        a2.triggered.connect(self.rotateBlockCW)

        ll = _img.ROTATE_LEFT_PNG.icon()
        a3 = menu.addAction(ll, "Rotate Block counter-clockwise")
        a3.triggered.connect(self.rotateBlockCCW)

        a4 = menu.addAction("Reset Rotation")
        a4.triggered.connect(self.resetRotation)

        b1 = menu.addAction("Print Rotation")
        b1.triggered.connect(self.printRotation)

        c1 = menu.addAction("Delete this Block")
        c1.triggered.connect(self.deleteBlockCom)

        menu.exec_(event.screenPos())

    def launchNotepadFile(self):
        self.logger.debug("Launching notpad")
        os.system("start notepad++ " + FILEPATH)

    def mouseDoubleClickEvent(self, event):  # pylint: disable=unused-argument
        if hasattr(self, "isTempering"):
            self.parent.parent().showTVentilDlg(self)
        elif self.name == "Pump":
            self.parent.parent().showPumpDlg(self)
        elif self.name in ("TeePiece", "WTap_main"):
            self.parent.parent().showBlockDlg(self)
        elif self.name in ["SPCnr", "DPCnr", "DPTee"]:
            self.parent.parent().showDoublePipeBlockDlg(self)
        else:
            self.parent.parent().showBlockDlg(self)
            if len(self.propertyFile) > 0:
                for files in self.propertyFile:
                    os.startfile(files, "open")

    def mouseReleaseEvent(self, event):
        # self.logger.debug("Released mouse over block")
        if self.oldPos is None:
            self.logger.debug("For Undo Framework: oldPos is None")
        else:
            if self.scenePos() != self.oldPos:
                self.logger.debug("Block was dragged")
                self.logger.debug("Old pos is" + str(self.oldPos))
                command = MoveCommand(self, self.oldPos, "Move BlockItem")
                self.parent.parent().parent().undoStack.push(command)
                self.oldPos = self.scenePos()

        super().mouseReleaseEvent(event)

    # Transform related
    def changeSize(self):
        self._positionLabel()

    def _positionLabel(self):
        width, height = self._getCappedWithAndHeight()
        rect = self.label.boundingRect()
        labelWidth = rect.width()
        labelPosX = (height - labelWidth) / 2
        self.label.setPos(labelPosX, width)

    def _getCappedWithAndHeight(self):
        width = self.w
        height = self.h
        height = max(height, 20)
        width = max(width, 40)
        return width, height

    def updateFlipStateH(self, state):
        self.flippedH = bool(state)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

        self.flippedHInt = 1 if self.flippedH else -1

        if self.flippedH:
            for i, inputPort in enumerate(self.inputs):
                distanceToMirrorAxis = self.w / 2.0 - self.origInputsPos[i][0]  # pylint:disable=unsubscriptable-object
                inputPort.setPos(
                    self.origInputsPos[i][0] + 2.0 * distanceToMirrorAxis,  # pylint: disable = unsubscriptable-object
                    inputPort.pos().y(),
                )

            for i, outPort in enumerate(self.outputs):
                distanceToMirrorAxis = self.w / 2.0 - self.origOutputsPos[i][0]  # pylint:disable=unsubscriptable-object
                outPort.setPos(
                    self.origOutputsPos[i][0] + 2.0 * distanceToMirrorAxis,  # pylint: disable = unsubscriptable-object
                    outPort.pos().y(),
                )

        else:
            for i, inputPort in enumerate(self.inputs):
                inputPort.setPos(self.origInputsPos[i][0],  # pylint: disable = unsubscriptable-object
                                 inputPort.pos().y())

            for i, outputPort in enumerate(self.outputs):
                outputPort.setPos(self.origOutputsPos[i][0],  # pylint: disable = unsubscriptable-object
                                  outputPort.pos().y())

    def updateFlipStateV(self, state):
        self.flippedV = bool(state)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

        self.flippedVInt = 1 if self.flippedV else -1

        if self.flippedV:
            for i, inputPort in enumerate(self.inputs):
                distanceToMirrorAxis = self.h / 2.0 - self.origInputsPos[i][1]  # pylint:disable=unsubscriptable-object
                inputPort.setPos(
                    inputPort.pos().x(),
                    self.origInputsPos[i][1] + 2.0 * distanceToMirrorAxis,  # pylint: disable = unsubscriptable-object
                )

            for i, outputPort in enumerate(self.outputs):
                distanceToMirrorAxis = self.h / 2.0 - self.origOutputsPos[i][1]  # pylint:disable=unsubscriptable-object
                outputPort.setPos(
                    outputPort.pos().x(),
                    self.origOutputsPos[i][1] + 2.0 * distanceToMirrorAxis,  # pylint: disable = unsubscriptable-object
                )

        else:
            for i, inputPort in enumerate(self.inputs):
                inputPort.setPos(inputPort.pos().x(),
                                 self.origInputsPos[i][1])  # pylint: disable = unsubscriptable-object

            for i, outputPort in enumerate(self.outputs):
                outputPort.setPos(outputPort.pos().x(),
                                  self.origOutputsPos[i][1])  # pylint: disable = unsubscriptable-object

    def updateSidesFlippedH(self):
        if self.rotationN % 2 == 0:
            for p in self.inputs:
                if p.side in (0, 2):
                    self.updateSide(p, 2)
            for p in self.outputs:
                if p.side in (0, 2):
                    self.updateSide(p, 2)
        if self.rotationN % 2 == 1:
            for p in self.inputs:
                if p.side in (1, 3):
                    self.updateSide(p, 2)
            for p in self.outputs:
                if p.side in (1, 3):
                    self.updateSide(p, 2)

    def updateSidesFlippedV(self):
        if self.rotationN % 2 == 1:
            for p in self.inputs:
                if p.side in (0, 2):
                    self.updateSide(p, 2)
            for p in self.outputs:
                if p.side in (0, 2):
                    self.updateSide(p, 2)
        if self.rotationN % 2 == 0:
            for p in self.inputs:
                if p.side in (1, 3):
                    self.updateSide(p, 2)
            for p in self.outputs:
                if p.side in (1, 3):
                    self.updateSide(p, 2)

    def updateSide(self, port, n):  # pylint: disable=no-self-use
        port.side = (port.side + n) % 4

    def rotateBlockCW(self):
        # Rotate block clockwise
        # self.setTransformOriginPoint(50, 50)
        # self.setTransformOriginPoint(self.w/2, self.h/2)
        self.setTransformOriginPoint(0, 0)
        self.setRotation((self.rotationN + 1) * 90)
        self.label.setRotation(-(self.rotationN + 1) * 90)
        self.rotationN += 1
        self.logger.debug("rotated by " + str(self.rotationN))

        for p in self.inputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, 1)

        for p in self.outputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, 1)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def rotateBlockToN(self, n):
        if n > 0:
            while self.rotationN != n:
                self.rotateBlockCW()
        if n < 0:
            while self.rotationN != n:
                self.rotateBlockCCW()

    def rotateBlockCCW(self):
        # Rotate block clockwise
        # self.setTransformOriginPoint(50, 50)
        self.setTransformOriginPoint(0, 0)
        self.setRotation((self.rotationN - 1) * 90)
        self.label.setRotation(-(self.rotationN - 1) * 90)
        self.rotationN -= 1
        self.logger.debug("rotated by " + str(self.rotationN))

        for p in self.inputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, -1)

        for p in self.outputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, -1)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def resetRotation(self):
        self.logger.debug("Resetting rotation...")
        self.setRotation(0)
        self.label.setRotation(0)

        for p in self.inputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, -self.rotationN)
            # self.logger.debug("Portside of port " + str(p) + " is " + str(p.portSide))

        for p in self.outputs:
            p.itemChange(27, p.scenePos())
            self.updateSide(p, -self.rotationN)
            # self.logger.debug("Portside of port " + str(p) + " is " + str(p.portSide))

        self.rotationN = 0

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def printRotation(self):
        self.logger.debug("Rotation is " + str(self.rotationN))

    # Deletion related
    def deleteConns(self):

        for p in self.inputs:
            while len(p.connectionList) > 0:
                p.connectionList[0].deleteConn()

        for p in self.outputs:
            while len(p.connectionList) > 0:
                p.connectionList[0].deleteConn()

    def deleteBlock(self):
        self.parent.parent().trnsysObj.remove(self)
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        if widgetToRemove:
            widgetToRemove.hide()

    def deleteBlockCom(self):
        self.parent.deleteBlockCom(self)

    def getConnections(self):
        """
        Get the connections from inputs and outputs of this block.
        Returns
        -------
        c : :obj:`List` of :obj:`BlockItem`
        """
        c = []
        for i in self.inputs:
            for cl in i.connectionList:
                c.append(cl)
        for o in self.outputs:
            for cl in o.connectionList:
                c.append(cl)
        return c

    # Scaling related
    def mousePressEvent(self, event):  # pylint: disable = unused-argument
        """
        Using try catch to avoid creating extra resizers.

        When an item is clicked on, it will check if a resizer already existed. If
        there exist a resizer, returns. else, creates one.

        Resizer will not be created for GenericBlock due to complications in the code.
        Resizer will not be created for storageTank as there's already a built in function for it in the storageTank
        dialog.

        Resizers are deleted inside mousePressEvent function inside GUI.py

        """
        self.logger.debug("Inside Block Item mouse click")

        self.isSelected = True
        if self.name in ("GenericBlock", "StorageTank"):
            return
        try:
            self.resizer
        except AttributeError:
            self.resizer = ResizerItem(self)  # pylint: disable = attribute-defined-outside-init
            self.resizer.setPos(self.w, self.h)
            self.resizer.itemChange(self.resizer.ItemPositionChange, self.resizer.pos())

    def setItemSize(self, w, h):
        self.logger.debug("Inside block item set item size")
        self.w, self.h = w, h

    def updateImage(self):
        self.logger.debug("Inside block item update image")

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

        if self.flippedH:
            self.updateFlipStateH(self.flippedH)

        if self.flippedV:
            self.updateFlipStateV(self.flippedV)

    def _getPixmap(self) -> QPixmap:
        imageAccessor = self._getImageAccessor()  # pylint: disable = assignment-from-no-return

        assert imageAccessor

        image = imageAccessor.image(width=self.w, height=self.h).mirrored(
            horizontal=self.flippedH, vertical=self.flippedV
        )
        pixmap = QPixmap(image)

        return pixmap

    def deleteResizer(self):
        try:
            self.resizer
        except AttributeError:
            self.logger.debug("No resizer")
        else:
            del self.resizer

    # AlignMode related
    def itemChange(self, change, value):
        # Snap grid excludes alignment

        if change == self.ItemPositionChange:
            if self.parent.parent().snapGrid:
                snapSize = self.parent.parent().snapSize
                self.logger.debug("itemchange")
                self.logger.debug(type(value))
                value = QPointF(value.x() - value.x() % snapSize, value.y() - value.y() % snapSize)
                return value
            if self.parent.parent().alignMode:
                if self.hasElementsInYBand():
                    return self.alignBlock(value)
            return value
        return super().itemChange(change, value)

    def alignBlock(self, value):
        for t in self.parent.parent().trnsysObj:
            if isinstance(t, BlockItem) and t is not self:
                if self.elementInYBand(t):
                    value = QPointF(self.pos().x(), t.pos().y())
                    self.parent.parent().alignYLineItem.setLine(
                        self.pos().x() + self.w / 2, t.pos().y(), t.pos().x() + t.w / 2, t.pos().y()
                    )

                    self.parent.parent().alignYLineItem.setVisible(True)

                    qtm = QTimer(self.parent.parent())
                    qtm.timeout.connect(self.timerfunc)
                    qtm.setSingleShot(True)
                    qtm.start(1000)

                    e = QMouseEvent(
                        QEvent.MouseButtonRelease,
                        self.pos(),
                        QtCore.Qt.NoButton,
                        QtCore.Qt.NoButton,
                        QtCore.Qt.NoModifier,
                    )
                    self.parent.mouseReleaseEvent(e)
                    self.parent.parent().alignMode = False

                if self.elementInXBand(t):
                    value = QPointF(t.pos().x(), self.pos().y())
                    self.parent.parent().alignXLineItem.setLine(
                        t.pos().x(), t.pos().y() + self.w / 2, t.pos().x(), self.pos().y() + t.w / 2
                    )

                    self.parent.parent().alignXLineItem.setVisible(True)

                    qtm = QTimer(self.parent.parent())
                    qtm.timeout.connect(self.timerfunc2)
                    qtm.setSingleShot(True)
                    qtm.start(1000)

                    e = QMouseEvent(
                        QEvent.MouseButtonRelease,
                        self.pos(),
                        QtCore.Qt.NoButton,
                        QtCore.Qt.NoButton,
                        QtCore.Qt.NoModifier,
                    )
                    self.parent.mouseReleaseEvent(e)
                    self.parent.parent().alignMode = False

        return value

    def timerfunc(self):
        self.parent.parent().alignYLineItem.setVisible(False)

    def timerfunc2(self):
        self.parent.parent().alignXLineItem.setVisible(False)

    def hasElementsInYBand(self):
        for t in self.parent.parent().trnsysObj:
            if isinstance(t, BlockItem):
                if self.elementInYBand(t):
                    return True

        return False

    def hasElementsInXBand(self):
        for t in self.parent.parent().trnsysObj:
            if isinstance(t, BlockItem):
                if self.elementInXBand(t):
                    return True

        return False

    def elementInYBand(self, t):
        eps = 50
        return self.scenePos().y() - eps <= t.scenePos().y() <= self.scenePos().y() + eps

    def elementInXBand(self, t):
        eps = 50
        return self.scenePos().x() - eps <= t.scenePos().x() <= self.scenePos().x() + eps

    def elementInY(self):
        for t in self.parent.parent().trnsysObj:
            if isinstance(t, BlockItem):
                if self.scenePos().y == t.scenePos().y():
                    return True
        return False

    def encode(self):
        portListInputs = []
        portListOutputs = []

        for inp in self.inputs:
            portListInputs.append(inp.id)
        for output in self.outputs:
            portListOutputs.append(output.id)

        blockPosition = (float(self.pos().x()), float(self.pos().y()))

        blockItemModel = BlockItemModel(
            self.name,
            self.displayName,
            blockPosition,
            self.id,
            self.trnsysId,
            portListInputs,  # pylint: disable = duplicate-code # 1
            portListOutputs,
            self.flippedH,
            self.flippedV,
            self.rotationN,
        )

        dictName = "Block-"
        return dictName, blockItemModel.to_dict()

    def decode(self, i, resBlockList):
        model = BlockItemModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.Id
        self.trnsysId = model.trnsysId

        if len(self.inputs) != len(model.portsIdsIn) or len(self.outputs) != len(model.portsIdsOut):
            temp = model.portsIdsIn
            model.portsIdsIn = model.portsIdsOut
            model.portsIdsOut = temp

        for index, inp in enumerate(self.inputs):
            inp.id = model.portsIdsIn[index]

        for index, out in enumerate(self.outputs):
            out.id = model.portsIdsOut[index]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        resBlockList.append(self)

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList,
                    **kwargs):  # pylint: disable=unused-argument
        self.setPos(
            float(i["BlockPosition"][0] + offset_x),
            float(i["BlockPosition"][1] + offset_y),
        )

        self.updateFlipStateH(i["FlippedH"])
        self.updateFlipStateV(i["FlippedV"])
        self.rotateBlockToN(i["RotationN"])

        for x, inputPort in enumerate(self.inputs):
            inputPort.id = i["PortsIDIn"][x]

        for x, outputPort in enumerate(self.outputs):
            outputPort.id = i["PortsIDOut"][x]

        resBlockList.append(self)

    # Export related
    def exportBlackBox(self):
        equation = []
        if (
                len(self.inputs + self.outputs) == 2
                and self.isVisible()
                and not isinstance(self.outputs[0], DoublePipePortItem)
        ):
            files = glob.glob(os.path.join(self.path, "**/*.ddck"), recursive=True)
            if not files:
                status = "noDdckFile"
            else:
                status = "noDdckEntry"
            lines = []
            for file in files:
                with open(file, "r") as infile:  # pylint: disable=unspecified-encoding
                    lines += infile.readlines()
            for i, line in enumerate(lines):
                if "output" in line.lower() and "to" in line.lower() and "hydraulic" in line.lower():
                    for j in range(i, len(lines) - i):
                        if lines[j][0] == "T":
                            outputT = lines[j].split("=")[0].replace(" ", "")
                            status = "success"
                            break
                    equation = ["T" + self.displayName + "=" + outputT]
                    break
        else:
            status = "noBlackBoxOutput"

        if status in ("noDdckFile", "noDdckEntry"):
            equation.append("T" + self.displayName + "=1")

        return status, equation

    def exportPumpOutlets(self):  # pylint: disable=no-self-use
        return "", 0

    def exportMassFlows(self):  # pylint: disable=no-self-use
        return "", 0

    def exportDivSetting1(self):  # pylint: disable=no-self-use
        return "", 0

    def exportDivSetting2(self, nUnit):  # pylint: disable=no-self-use
        return "", nUnit

    def exportPipeAndTeeTypesForTemp(self, startingUnit: int) -> _tp.Tuple[str, int]:  # pylint: disable=no-self-use
        return "", startingUnit

    def getTemperatureVariableName(self, portItem: SinglePipePortItem) -> str:  # pylint: disable=unused-argument
        return f"T{self.displayName}"

    def getFlowSolverParametersId(self, portItem: SinglePipePortItem) -> int:  # pylint: disable=unused-argument
        return self.trnsysId

    def assignIDsToUninitializedValuesAfterJsonFormatMigration(self, generator: _id.IdGenerator) -> None:
        pass

    def deleteLoadedFile(self):
        for items in self.loadedFiles:
            try:
                self.parent.parent().fileList.remove(str(items))
            except ValueError:
                self.logger.debug("File already deleted from file list.")
                self.logger.debug("filelist:", self.parent.parent().fileList)
